import json
import pathlib
import tempfile

from idem.exec.init import ExecReturn

__func_alias__ = {"exec_": "exec"}


async def sls(hub) -> int:
    """
    Execute the cli routine to run states
    """
    sls_sources = list(hub.OPT.idem.sls_sources)
    param_sources = list(hub.OPT.idem.param_sources)
    if hub.OPT.idem.tree:
        tree = f"file://{hub.OPT.idem.tree}"
        hub.log.debug(f"Added tree to sls and param sources: {tree}")
        sls_sources.append(tree)
        param_sources.append(tree)
    src = hub.idem.sls_source.init.get_refs(sources=sls_sources, refs=hub.OPT.idem.sls)
    param = hub.idem.sls_source.param.get_refs(
        sources=param_sources, refs=hub.OPT.idem.params
    )
    name = hub.idem.RUN_NAME
    cache_dir = hub.OPT.idem.cache_dir

    # Get the acct information for the enforced state manager
    context_manager = hub.idem.managed.context(
        run_name=name,
        cache_dir=cache_dir,
        esm_plugin=hub.OPT.idem.esm_plugin,
        esm_profile=hub.OPT.idem.esm_profile,
        acct_file=hub.OPT.acct.acct_file,
        acct_key=hub.OPT.acct.acct_key,
        serial_plugin=hub.OPT.idem.esm_serial_plugin,
    )

    # Prepare the config dir as a backup sls source
    if hub.OPT.idem.config:
        config_dir = (
            f"file://{pathlib.Path(hub.OPT.idem.config).parent.resolve().absolute()}"
        )

        # Always use the config dir as a backup sls source
        hub.log.debug(f"Using config dir as sls source: {config_dir}")
        src["sls_sources"].append(config_dir)
        hub.log.debug(f"Using config dir as param source: {config_dir}")
        param["param_sources"].append(config_dir)

    # Run idem!
    try:
        async with context_manager as state:
            await hub.idem.state.apply(
                name=name,
                sls_sources=src["sls_sources"],
                render=hub.OPT.idem.render,
                runtime=hub.OPT.idem.runtime,
                subs=["states"],
                cache_dir=cache_dir,
                sls=src["sls"],
                target=hub.OPT.idem.target,
                test=hub.OPT.idem.test,
                invert_state=hub.OPT.idem.invert,
                acct_file=hub.OPT.acct.acct_file,
                acct_key=hub.OPT.acct.acct_key,
                acct_profile=hub.OPT.idem.acct_profile,
                managed_state=state,
                param_sources=param["param_sources"],
                params=param["params"],
            )

            errors = hub.idem.RUNS[name]["errors"]
            if errors:
                display = hub.output.nested.display(errors)
                print(display)
                # Return a non-zero error code
                return len(errors)

            # Reconciliation loop
            # Skip reconciliation during 'test'
            if hub.OPT.idem.test is False:
                await hub.reconcile.init.run(
                    plugin=hub.OPT.idem.reconciler,
                    pending_plugin=hub.OPT.idem.pending,
                    name=name,
                    apply_kwargs=dict(
                        sls_sources=src["sls_sources"],
                        render=hub.OPT.idem.render,
                        runtime=hub.OPT.idem.runtime,
                        cache_dir=hub.OPT.idem.cache_dir,
                        sls=src["sls"],
                        test=hub.OPT.idem.test,
                        acct_file=hub.OPT.acct.acct_file,
                        acct_key=hub.OPT.acct.acct_key,
                        acct_profile=hub.OPT.idem.acct_profile,
                        subs=["states"],
                        managed_state=state,
                    ),
                )
            else:
                # FINISHED event is triggered by the reconciliation.
                # Triggering the FINISHED event here since reconciliation is
                # skipped in 'test' mode.
                hub.idem.state.update_status(name, hub.idem.state.Status.FINISHED)
    except Exception as e:
        display = hub.output.nested.display(str(e))
        print(display)
        # Return a non-zero error code
        return 1

    running = hub.idem.RUNS[name]["running"]

    output = hub.OPT.rend.output or "state"
    grouped_data = hub.group.init.apply(running)
    display = hub.output[output].display(grouped_data)
    print(display)
    return 0


async def exec_(hub) -> int:
    exec_path = hub.OPT.idem.exec_func
    exec_args = hub.OPT.idem.exec_args
    if not exec_path.startswith("exec"):
        exec_path = f"exec.{exec_path}"
    args = []
    kwargs = {}
    for arg in exec_args:
        if isinstance(arg, dict):
            kwargs.update(arg)
        else:
            args.append(arg)
    ret = await hub.idem.ex.run(
        path=exec_path,
        args=args,
        kwargs=kwargs,
        acct_file=hub.OPT.acct.acct_file,
        acct_key=hub.OPT.acct.acct_key,
        acct_profile=hub.OPT.idem.acct_profile,
    )

    output = hub.OPT.rend.output or "exec"
    display = hub.output[output].display(ret)
    print(display)

    if isinstance(ret, ExecReturn):
        return int(not ret.result)

    return 1


async def desc(hub) -> int:
    state_path = hub.OPT.idem.desc_glob
    try:
        ret = await hub.idem.describe.run(
            state_path,
            hub.OPT.acct.acct_file,
            hub.OPT.acct.acct_key,
            hub.OPT.idem.acct_profile,
            progress=hub.OPT.idem.progress,
            hard_fail=hub.OPT.idem.hard_fail,
            search_path=hub.OPT.idem.filter,
        )
    except Exception as e:
        display = hub.output.nested.display(str(e))
        print(display)
        # Return a non-zero error code
        return 1

    output = hub.OPT.rend.output or "yaml"
    display = hub.output[output].display(ret)
    print(display)
    return 0


async def validate(hub) -> int:
    """
    Execute the cli routine to validate states
    """
    sls_sources = hub.OPT.idem.sls_sources
    if hub.OPT.idem.tree:
        sls_sources.append(f"file://{hub.OPT.idem.tree}")
    src = hub.idem.sls_source.init.get_refs(sources=sls_sources, refs=hub.OPT.idem.sls)
    name = hub.OPT.idem.run_name
    await hub.idem.state.validate(
        name=name,
        sls_sources=src["sls_sources"],
        render=hub.OPT.idem.render,
        runtime=hub.OPT.idem.runtime,
        subs=["states"],
        cache_dir=hub.OPT.idem.cache_dir,
        sls=src["sls"],
        test=hub.OPT.idem.test,
        invert_state=hub.OPT.idem.invert,
        acct_file=hub.OPT.acct.acct_file,
        acct_key=hub.OPT.acct.acct_key,
        acct_profile=hub.OPT.idem.acct_profile,
    )

    errors = hub.idem.RUNS[name]["errors"]
    if errors:
        display = hub.output.nested.display(errors)
        print(display)
        # Return a non-zero error code
        return len(errors)

    ret = {
        "high": hub.idem.RUNS[name]["high"],
        "low": hub.idem.RUNS[name]["low"],
        "meta": hub.idem.RUNS[name]["meta"],
        "parameters": hub.idem.RUNS[name]["parameters"],
        "warnings": hub.idem.RUNS[name]["warnings"],
    }
    output = hub.OPT.rend.output or "nested"
    display = hub.output[output].display(ret)
    print(display)
    return 0


async def refresh(hub) -> int:
    """
    Update enforced state management with described resources.
    Run "describe" for the given path, then run `idem state --test` on it's output.
    This brings it in to the enforced state management.
    Nothing should be changed on the resources after this command.
    """
    state_path = hub.OPT.idem.desc_glob
    run_name = hub.idem.RUN_NAME
    cache_dir = hub.OPT.idem.cache_dir
    output = hub.OPT.rend.output or "state"

    # Generate an sls file based on "describe"
    ret = await hub.idem.describe.run(
        state_path,
        hub.OPT.acct.acct_file,
        hub.OPT.acct.acct_key,
        hub.OPT.idem.acct_profile,
        progress=hub.OPT.idem.progress,
        hard_fail=hub.OPT.idem.hard_fail,
        search_path=None,
    )
    if not ret:
        raise ValueError(f"No descriptions available for the given path: {state_path}")

    # Get the acct information for the enforced state manager
    context_manager = hub.idem.managed.context(
        run_name=run_name,
        cache_dir=cache_dir,
        esm_plugin=hub.OPT.idem.esm_plugin,
        esm_profile=hub.OPT.idem.esm_profile,
        acct_file=hub.OPT.acct.acct_file,
        acct_key=hub.OPT.acct.acct_key,
        serial_plugin=hub.OPT.idem.esm_serial_plugin,
    )
    try:
        async with context_manager as state:
            # Write the describe output to a file
            with tempfile.NamedTemporaryFile(suffix=".sls", delete=True) as fh:
                path = pathlib.Path(fh.name)
                display = hub.output.json.display(ret)
                fh.write(display.encode())
                fh.flush()

                # Apply the state from the describe file
                await hub.idem.state.apply(
                    name=run_name,
                    sls_sources=[f"file://{path.parent}"],
                    render="json",
                    runtime="parallel",
                    subs=["states"],
                    cache_dir=cache_dir,
                    sls=[path.stem],
                    test=True,
                    acct_file=hub.OPT.acct.acct_file,
                    acct_key=hub.OPT.acct.acct_key,
                    acct_profile=hub.OPT.idem.acct_profile,
                    managed_state=state,
                )
    except Exception as e:
        display = hub.output.nested.display(str(e))
        print(display)
        # Return a non-zero error code
        return 1

    # Report Errors
    errors = hub.idem.RUNS[run_name]["errors"]
    if errors:
        display = hub.output.nested.display(errors)
        print(display)
        # Return a non-zero error code
        return len(errors)

    # If something changed, which it shouldn't, it will show up now
    running = hub.idem.RUNS[run_name]["running"]
    # Get all the describe states that reported changes
    changed = {k: v for k, v in running.items() if v.get("changes")}
    display = hub.output[output].display(changed)
    print(display)
    if changed:
        hub.log.error(f"Changes were made by refresh on path: {state_path}")
    return len(changed)


async def restore(hub) -> int:
    """
    Restore the centralized state management context from a local json file
    """
    name = hub.idem.RUN_NAME
    cache_dir = hub.OPT.idem.cache_dir
    esm_plugin = hub.OPT.idem.esm_plugin

    with open(hub.OPT.idem.esm_cache_file) as fh:
        restore_data = json.load(fh)

    # Get the acct information for the centralized state manager
    context_manager = hub.idem.managed.context(
        run_name=name,
        cache_dir=cache_dir,
        esm_plugin=esm_plugin,
        esm_profile=hub.OPT.idem.esm_profile,
        acct_file=hub.OPT.acct.acct_file,
        acct_key=hub.OPT.acct.acct_key,
        serial_plugin=hub.OPT.idem.esm_serial_plugin,
    )
    try:
        async with context_manager as state:
            state.update(restore_data)
    except Exception as e:
        display = hub.output.nested.display(str(e))
        print(display)
        # Return a non-zero error code
        return 1

    return 0


async def doc(hub) -> int:
    """
    Get documentation for functions under the given reference
    """
    # Get all the documentation from the hub as a single tree
    tree = await hub.tree.init.traverse()

    if hub.OPT.pop_tree.ref:
        # Parse the tree for the named reference
        tree = hub.tree.ref.get(tree, hub.OPT.pop_tree.ref)

        if not tree:
            hub.log.error(
                f"Reference does not exist on the hub: {hub.OPT.pop_tree.ref}"
            )
            return 1

        # Flatten the tree so that each full reference is a top-level key
        ret = hub.tree.ref.list(tree)

        # Get all the references under the named reference from the tree
        result = {}
        for ref, docs in ret.items():
            if ref.startswith(hub.OPT.pop_tree.ref) and "parameters" in docs:
                result[ref] = docs

        if not result:
            result = {hub.OPT.pop_tree.ref: ret[hub.OPT.pop_tree.ref]}
    else:
        # Flatten the tree and return every function reference
        result = {}
        ret = hub.tree.ref.list(tree)
        for ref, docs in ret.items():
            if "parameters" in docs:
                result[ref] = docs

    outputter = hub.OPT.rend.output or "nested"

    print(hub.output[outputter].display(result))
    return 0



def get(
    hub,
    wait_in_seconds: int = None,
    multiplier: int = None,
    run_count: int = 0,
    **kwargs,
):
    # return wait time in seconds
    # run_count is an optional argument and is the number of retries already occurred
    # wait time is calculated as wait_in_seconds * (multiplier ^ run_count)
    # For example: wait_in_seconds = 2, multiplier 10,
    # exponential wait times are: 2, 20, 200...
    #
    # Define in the state:
    # __reconcile_wait__ = { "exponential": {"wait_in_seconds": 2, "multiplier": 10}}

    if (
        wait_in_seconds is None
        or multiplier is None
        or wait_in_seconds == 0
        or multiplier == 0
    ):
        raise ValueError(
            "Expecting wait_in_seconds and multiplier to be specified for exponential backoff."
        )

    val = wait_in_seconds
    if run_count:
        val = val * (multiplier**run_count)

    hub.log.debug(f"Exponential value: {val}")
    return val



import copy
import fnmatch
import inspect
import warnings
from typing import Any
from typing import Dict
from typing import List


def gen_chunk_func_tag(hub, chunk):
    """
    Generate the unique tag used to track the execution of the chunk
    """
    if "sls_run_id" in chunk:
        return _gen_sls_run_tag(chunk, f'{_gen_tag(chunk)}{chunk["fun"]}')

    return f'{_gen_tag(chunk)}{chunk["fun"]}'


def gen_chunk_esm_tag(hub, chunk: Dict[str, str]) -> str:
    """
    Generate the unique tag used to track the execution of the chunk in the esm
    """
    if "sls_run_id" in chunk:
        return _gen_sls_run_tag(chunk, _gen_tag(chunk))

    return _gen_tag(chunk)


def _gen_sls_run_tag(chunk, tag):
    sls_run_tag = ""
    # As chuck has an attribute sls_run_id, it indicates this chunk is part of a sls.run.
    # so add the parent state. In this case parent state is sls
    sls_run_tag += "sls"

    # Add id of the chunk
    # This Identifies the tag uniquely. <sls.run name>:<Id of the resource>
    sls_run_tag += f'_|-{chunk["__id__"]}'

    # Add the rest of the tag as we do for a normal run
    sls_run_tag += f"_|-{tag}"

    return sls_run_tag


def _gen_tag(chunk: Dict[str, str]) -> str:
    """
    Generate the unique tag used to track the execution of the chunk

    This will be used at a module level, agnostic to function names
    """
    tag = f'{chunk["state"]}_|-{chunk["__id__"]}_|-{chunk["name"]}_|-'

    # if `name_prefix` is present and `name` is formed with `name_prefix`, then create tag based on `name_prefix`
    if chunk.get("name_prefix") and chunk.get("name_prefix") in chunk["name"]:
        tag = f'{chunk["state"]}_|-{chunk["__id__"]}_|-{chunk["name_prefix"]}_|-'

    return tag


def get_sls_run_chunks(hub, low, state, sls_run_id, path_args, from_esm=False):
    """
    Search in the low state for the chunk with the given state, sls_run_id and path_args
    we find the exact chunk inside group of sls_files in a sls.run with name sls_run_id.
    Example-: if we have path_args = [{'cloud2:state2:property_path': 'new_state:key'}]
    we will find the chunk in low data with state cloud2 and name state2 and sls_run of chunk
    should match sls_run_id and updated path_args would be [{property_path:'new_state:key'}]


    state4:
     cloud1.state_def:
      - parameters:
            arg1: ${sls:state1:sls:state2:cloud2:state3:property_path}

    In the above example the actual state or chunk is cloud2:state3:property_path which is part of a sls.run
    named state2 and state2 is part of another sls.run named state1. To find the exact chunk in low data we need to
    split and process the requisite recursively till we reach the exact chunk.

    In first run sls_run_id=state1 and path_args = [{'sls:state2:cloud2:state3:property_path': 'new_state:key'}]
    After first run we find chunks with name=state2 and state=sls, since state2 is also a sls.run we did not reach the exact chunk.

    In the second run sls_run_id=state1:sls:state2 and path_args = [{'cloud2:state2:property_path': 'new_state:key'}]
    In second run we search for state=cloud2 and name = state3 and sls_run_id=state1:sls:state2
    which indicates chunk is part of state2 and state2 should be part of state1

    the chunk we found after second un is not a sls.run. we reached the exact chunk,
    so return that chunk and updated_args (the value being referred in chunk)
    """
    chunks_found = []
    new_path_args = []
    for path_arg in path_args:
        name_def = next(iter(path_arg))
        name_def_list = name_def.split(":")
        if len(name_def_list) > 1:
            new_state = name_def_list.pop(0)
            new_name = name_def_list.pop(0)

            if from_esm:
                original_chunks = hub.tool.idem.esm.get_chunks_from_esm(
                    new_state, new_name, sls_run_id
                )
            else:
                original_chunks = hub.idem.tools.get_chunks(
                    low, new_state, sls_run_id + ":" + new_name
                )

            # check if chunks belong to same run else filter them
            original_chunks = [
                original_chunk
                for original_chunk in original_chunks
                if original_chunk["sls_run_id"] == sls_run_id
            ]
            if original_chunks:
                original_chunk = original_chunks[0]
                # After finding the exact chunk we need to check if the chunk is sls.run
                # if chunk is sls.run this means we have a nested sls.run, and we did not reach the exact chunk.
                # so we need to find the chunk recursively till we get a non sls.run chunk.
                # sls_run_id in the chunk for nested module will be <name of outer sls.run><hub.states.sls.SLS_RUN_SEPARATOR><name of inner sls.run>
                # so we call the method recursively with updated sls_run name, new state and new_path_args
                if (
                    original_chunk["state"] == "sls"
                    and original_chunk["fun"] == "run"
                    and len(name_def_list) > 1
                ):
                    hub.log.debug(
                        f"nested sls.run found. calling the function get_sls_run_chunks with updated sls_run_id {sls_run_id + hub.states.sls.SLS_RUN_SEPARATOR + new_name}"
                    )
                    new_arg_key = ":".join(name_def_list)
                    new_path_args.append({new_arg_key: path_arg[name_def]})
                    return hub.idem.tools.get_sls_run_chunks(
                        low,
                        new_state,
                        sls_run_id + hub.states.sls.SLS_RUN_SEPARATOR + new_name,
                        new_path_args,
                        from_esm,
                    )
            chunks_found.extend(original_chunks)
            new_arg_key = ":".join(name_def_list)
            new_path_args.append({new_arg_key: path_arg[name_def]})
        else:
            if from_esm:
                chunks_found = hub.tool.idem.esm.get_chunks_from_esm(
                    state, sls_run_id, sls_run_id
                )
            else:
                chunks_found = hub.idem.tools.get_chunks(low, state, sls_run_id)

            new_path_args = path_args
    return chunks_found[0] if chunks_found else None, new_path_args


def get_chunks(hub, low, state, name):
    """
    Search in the low state for the chunk with the given designation
    """
    rets = []
    for chunk in low:
        if state == "sls":
            if fnmatch.fnmatch(chunk["__sls__"], name):
                rets.append(chunk)
                continue
        if state == chunk["state"]:
            if fnmatch.fnmatch(chunk["name"], name) or fnmatch.fnmatch(
                chunk["__id__"], name
            ):
                rets.append(chunk)
    return rets


def find_name(hub, name, state, high):
    """
    Scan high data for the id referencing the given name and return a list of (IDs, state) tuples that match
    Note: if `state` is sls, then we are looking for all IDs that match the given SLS
    """
    ext_id = []
    if name in high:
        ext_id.append((name, state))
    # if we are requiring an entire SLS, then we need to add ourselves to everything in that SLS
    elif state == "sls":
        for nid, item in high.items():
            if item["__sls__"] == name:
                ext_id.append((nid, next(iter(item))))
    # otherwise we are requiring a single state, lets find it
    else:
        # We need to scan for the name
        for nid in high:
            if state in high[nid]:
                if isinstance(high[nid][state], list):
                    for arg in high[nid][state]:
                        if not isinstance(arg, dict):
                            continue
                        if len(arg) != 1:
                            continue
                        if arg[next(iter(arg))] == name:
                            ext_id.append((nid, state))
    return ext_id


def format_call(
    hub,
    fun,
    data,
    ignore_changes: List = None,
    expected_extra_kws=(),
    enforced_state: Dict[str, Any] = None,
):
    """
    Build the required arguments and keyword arguments required for the passed
    function.
    :param fun: The function to get the argspec from
    :param data: A dictionary containing the required data to build the
                 arguments and keyword arguments.
    :param ignore_changes: A list of path of parameters that will be assigned to None value
                           to ignore being updated in present().
    :param expected_extra_kws: Any expected extra keyword argument names which
                               should not trigger a :ref:`SaltInvocationError`
    :param enforced_state: A dictionary with the parameters from a previous run
    :returns: Function keyword arguments.
    """
    enforced_state = enforced_state or {}

    args = []
    kwargs = {}
    keywords = False

    sig = fun.signature
    for name, param in sig.parameters.items():
        if name == "hub":
            continue
        elif param.kind.name == "POSITIONAL_OR_KEYWORD":
            if param.default is inspect._empty:
                if name in enforced_state:
                    # get it from enforced state before state
                    kwargs[name] = enforced_state[name]
                else:
                    # Warn about required argument that is not stored in ESM
                    if enforced_state and name != "ctx":
                        msg = (
                            f"Function {fun.ref}.{fun.func.__name__} argument '{name}' is required, "
                            f"but is not found in ESM cache for {gen_chunk_func_tag(hub, data)}"
                        )
                        warnings.warn(msg, RuntimeWarning)
                        hub.log.warning(msg)
                    args.append(name)
            else:
                kwargs[name] = enforced_state.get(name, param.default)
                # Validate a boolean value for arguments of type boolean
                if isinstance(param.default, bool):
                    if data.get(name) and not isinstance(data.get(name), bool):
                        raise TypeError(
                            f"{fun.ref}.{fun.func.__name__} is expecting a boolean value for '{name}' but got '{data.get(name)}'"
                        )
        elif param.kind.name == "KEYWORD_ONLY":
            # get it from enforced state before default
            kwargs[name] = enforced_state.get(name, param.default)
        elif param.kind.name == "VAR_KEYWORD":
            keywords = True

    # Since we WILL be changing the data dictionary, let's change a copy of it
    data = data.copy()
    is_existing_resource = bool(enforced_state) or bool(data.get("resource_id"))

    missing_args = []

    for key in kwargs:
        try:
            # Do not override enforced_state value with None,
            # unless no value is set
            val = data.pop(key)
            if val is not None or kwargs[key] is inspect._empty:
                kwargs[key] = val

            # if the execution flow is for recreating a resource,
            # then we should override enforced_state value with None for resource_id
            if key == "resource_id" and data.get("recreation_flow", False):
                kwargs[key] = val
        except KeyError:
            # Let's leave the default value in place
            pass

    # If ignore_changes contains parameters, then we try to assign those parameters with None to skip updating.
    # For resource recreation flow, we should not nullify the parameters in ignore_changes.
    if (
        ignore_changes
        and is_existing_resource
        and not data.get("recreation_flow", False)
    ):
        hub.idem.tool.ignore_changes.ignore_parameter_changes(
            ignore_changes=ignore_changes,
            params=kwargs,
            param_signatures=sig.parameters,
        )

    while args:
        # Every function takes ctx as an argument.
        # All others are required function arguments that do not exist in ESM
        arg = args.pop(0)
        try:
            kwargs[arg] = data.pop(arg)
        except KeyError:
            missing_args.append(arg)

    if missing_args:
        raise ValueError(
            f"{fun.ref}.{fun.func.__name__} is missing required argument(s): {', '.join(missing_args)}"
        )

    if keywords:
        # The function accepts **kwargs, any non expected extra keyword
        # arguments will made available.
        for key, value in data.items():
            if key in expected_extra_kws:
                continue
            kwargs[key] = value

        # No need to check for extra keyword arguments since they are all
        # **kwargs now. Return
        return kwargs

    # Did not return yet? Lets gather any remaining and unexpected keyword
    # arguments
    extra = {}
    for key, value in data.items():
        if key in expected_extra_kws:
            continue
        extra[key] = copy.deepcopy(value)

    if extra:
        # Found unexpected keyword arguments, raise an error to the user
        if len(extra) == 1:
            msg = "'{0[0]}' is an invalid keyword argument for '{1}'".format(
                list(extra.keys()), f"{fun.__module__}.{fun.__name__}"
            )
        else:
            msg = "{} and '{}' are invalid keyword arguments for '{}'".format(
                ", ".join([f"'{e}'" for e in extra][:-1]),
                list(extra.keys())[-1],
                f"{fun.__module__}.{fun.__name__}",
            )
        warnings.warn(msg, RuntimeWarning)
        hub.log.warning(msg)
    return kwargs


def ishashable(hub, obj):
    """
    A simple test to verify if a given object is hashable and can therefore
    be used as a key in a dict
    """
    try:
        hash(obj)
    except TypeError:
        return False
    return True


def iter_high(hub, high):
    """
    Take a high-state structure and iterate over it yielding the elements down to the
    execution args
    Yields (id_, body, state, run, arg)
    """
    for id_, body in high.items():
        if not isinstance(body, dict):
            continue
        for state, run in body.items():
            if state.startswith("__"):
                continue
            for arg in run:
                yield id_, body, state, run, arg


def iter_high_leaf_args(hub, high):
    """
    Take a high-state structure and iterate over it yielding the elements down to the
    execution args
    Yields (id_, body, state, run, arg)
    """
    for id_, body in high.items():
        if not isinstance(body, dict):
            continue
        for state, run in body.items():
            if state.startswith("__"):
                continue
            for arg in run:
                yield from _iter_arg_chain(id_, body, state, run, "", arg)


def _iter_arg_chain(id_, body, state, run, arg_key_def, arg):
    if isinstance(arg, dict):
        for arg_key, arg_value in arg.items():
            # Escape dictionary references in the key definition to skip evaluating by arg_bind resolver.
            arg_key = arg_key.replace("[", "[\\")
            if arg_key_def:
                current_arg_key_def = arg_key_def + ":" + arg_key
            else:
                current_arg_key_def = arg_key

            if isinstance(arg_value, dict):
                yield from _iter_arg_chain(
                    id_, body, state, run, current_arg_key_def, arg_value
                )
            elif isinstance(arg_value, list):
                for index, arg_value_item in enumerate(arg_value):
                    yield from _iter_arg_chain(
                        id_,
                        body,
                        state,
                        run,
                        current_arg_key_def + f"[{index}]",
                        arg_value_item,
                    )
            else:
                yield id_, body, state, run, current_arg_key_def, arg_value
    elif isinstance(arg, list):
        for index, arg_value_item in enumerate(arg):
            yield from _iter_arg_chain(
                id_, body, state, run, arg_key_def + f"[{index}]", arg_value_item
            )
    else:
        yield id_, body, state, run, arg_key_def, arg


def get_enforced_state(hub, chunk, managed_state):
    tag = hub.idem.tools.gen_chunk_func_tag(chunk)
    esm_tag = hub.idem.tools.gen_chunk_esm_tag(chunk)
    new_chunk = {**chunk, "__id__": f"{chunk['__id__']}_create_new"}
    new_tag = hub.idem.tools.gen_chunk_func_tag(new_chunk)
    new_esm_tag = hub.idem.tools.gen_chunk_esm_tag(new_chunk)

    enforced_state = managed_state.get(new_esm_tag) or managed_state.get(new_tag)

    if not enforced_state:
        enforced_state = managed_state.get(esm_tag) or managed_state.get(tag)

    return enforced_state



import re


async def resolve(hub, arg_bind: str):
    """
    Parse arg_bind template ${cloud:state:attribute} and resolve the arg-bind template into a value
    using the hub's RUNNING data.
    """

    if arg_bind:
        result = re.search(r"\${([^${}]*)}", arg_bind)
        arg_bind_expr = result.group(1)

    resolved_data = None

    if arg_bind_expr:

        (
            resolved_data,
            is_data_available,
        ) = await hub.tool.idem.arg_bind_utils.find_arg_reference_data(arg_bind_expr)

        if not is_data_available:
            raise ValueError(
                f"Arg_bind template '{arg_bind}' could not be resolved to any value"
            )

    return resolved_data



import asyncio
import warnings
from typing import Any
from typing import Dict
from typing import List

import dict_tools.differ as differ


def sig_present(hub, ctx, name: str, *args, **kwargs):
    ...


def sig_absent(hub, ctx, name: str, *args, **kwargs):
    ...


async def sig_describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    ...


def post_present(hub, ctx):
    return _create_state_return(hub, ctx)


def post_absent(hub, ctx):
    return _create_state_return(hub, ctx)


def _create_state_return(hub, ctx):
    """
    Conform the output of every state return to this format.
    Valid state modules must return a dictionary with these keys
    Validate the output of every state return format.
    """
    if not ctx.ret:
        raise ValueError(f"No return from state")
    if (
        "old_state" in ctx.ret
        and "new_state" in ctx.ret
        and not ctx.ret.get("changes", None)
    ):
        old_state = ctx.ret.get("old_state")
        new_state = ctx.ret.get("new_state")
        ctx.ret["changes"] = differ.deep_diff(
            old_state if old_state else dict(), new_state if new_state else dict()
        )
    elif "changes" not in ctx.ret:
        hub.log.error(
            "States that implement 'resource' contract must either return 'changes' "
            "or a combination of 'old_state' and 'new_state'."
        )
    else:
        # TODO Raise an error here instead of a warning
        warnings.warn(
            f"It is required to return both 'old_state' and 'new_state' for the 'resource' contract.",
            DeprecationWarning,
        )
    # Prevent any sensitive data showing in "changes", which will be outputted to console.
    if "changes" in ctx.ret:
        arguments = ctx.get_arguments()
        if "run_name" in arguments["ctx"] and "tag" in arguments["ctx"]:
            run_name = arguments["ctx"]["run_name"]
            tag = arguments["ctx"]["tag"]
            if (
                "sensitive" in hub.idem.RUNS[run_name]
                and tag in hub.idem.RUNS[run_name]["sensitive"]
            ):
                # Put data-sanitization here because this is the closest to the Idem state module return.
                # However, in general, we should avoid adding too much functionalities in the contract's "post" signature.
                # Instead, the recommendation is to have a "post-rule" plugin to run after Idem state runs.
                if ctx.ret["changes"] and ctx.ret["changes"].get("new"):
                    for sensitive_data in hub.idem.RUNS[run_name]["sensitive"][tag]:
                        ctx.ret["changes"]["new"].pop(sensitive_data, None)
                if ctx.ret["changes"] and ctx.ret["changes"].get("old"):
                    for sensitive_data in hub.idem.RUNS[run_name]["sensitive"][tag]:
                        ctx.ret["changes"]["old"].pop(sensitive_data, None)

    try:
        return {
            "changes": ctx.ret["changes"],
            "comment": ctx.ret["comment"],
            "name": ctx.ret["name"],
            "result": ctx.ret["result"],
            "old_state": ctx.ret.get("old_state", None),
            "new_state": ctx.ret.get("new_state", None),
            "force_save": ctx.ret.get("force_save", False),
            "rerun_data": ctx.ret.get("rerun_data", None),
        }
    except KeyError:
        hub.log.error(f"Improperly formatted state return: {ctx.ref}")
        raise


def _verify_describe(hub, ret: Dict[str, Dict[str, Any]]):
    """
    Verify that the return value looks like
    {
        state_name: { path.present: [{},...] }
    }
    """
    for present_state in ret.values():
        for state_path, state_data in present_state.items():
            assert isinstance(
                state_data, List
            ), "State information should be formatted as a list"
            for item in state_data:
                assert isinstance(item, Dict), "Each item in the list should be a dict"
    return ret


async def _averify_describe(hub, ret):
    """
    Return a coroutine to a function that is expecting a coroutine
    """
    return _verify_describe(hub, await ret)


def pre_describe(hub, ctx):
    if hasattr(hub, "SUBPARSER"):
        if hub.SUBPARSER not in ("describe", "refresh"):
            raise ReferenceError(
                "'describe' functions should only be called by the describe subcommand."
                f"Please refer to the {ctx.ref} plugin documentation for 'search' functionality."
            )


def post_describe(hub, ctx):
    if asyncio.iscoroutine(ctx.ret):
        return _averify_describe(hub, ctx.ret)
    else:
        return _verify_describe(hub, ctx.ret)



from typing import Any
from typing import Dict


def run(
    hub,
    seq: Dict[int, Dict[str, Any]],
    low: Dict[str, Any],
    running: Dict[str, Any],
    options: Dict[str, Any],
) -> Dict[int, Dict[str, Any]]:
    """
    Process the requisites from ESM.
    A requisite can reference a state that is not present in the current run, but exists in ESM.
    This can happen when only subset of the resources are re-executed after the ESM was populated with
    Idem execution of the full set of the resources.
    In this case we want to look-up referenced state from the ESM for arg_bind and require requisites
    This function will check if requisites are present in this run
    If its present we do not do anything as requisites from the current
    run will already be added.
    If the requisite is not present in this run we check the ESM.
    if the resource is already run previously, and it exists in ESM,
    we prepare the requisite and add to the resource requisites.
    This applies to: arg_bind, require

    This function will only add requisites to resources if they are not present in current run and exists in ESM
    This will not modify the current running resources nor this will send any events.
    We only read a state from ESM to add to requisites. The state is not written back to ESM nor added to current run.
    """
    esm_enabled_req_types = ["arg_bind", "require"]
    for ind, data in seq.items():
        req_types_in_chunk = [
            req for req in esm_enabled_req_types if req in data["chunk"]
        ]
        if not req_types_in_chunk:
            continue
        chunk = data["chunk"]
        reqs = {}
        for type in req_types_in_chunk:
            reqs[type] = chunk.get(type)
        for req_type, rdefs in reqs.items():
            for rdef in rdefs:
                if not isinstance(rdef, dict):
                    data["errors"].append(f"{rdef} should be dictionary")
                    continue
                state = next(iter(rdef))
                if hasattr(hub.states, state):
                    if getattr(hub.states[state], "SKIP_ESM", False):
                        # This state doesn't use ESM
                        continue
                if isinstance(rdef[state], list):
                    name_defs = rdef[state]
                else:
                    name_defs = [{rdef[state]: []}]

                for name_def in name_defs:
                    if not isinstance(name_def, dict):
                        data["errors"].append(f"{name_def} should be dictionary")
                        continue
                    name = next(iter(name_def))
                    args = name_def[name]
                    r_chunks = hub.idem.tools.get_chunks(low, state, name)
                    if not r_chunks:
                        # If r_chunks is not found in current run check if its present in ESM
                        hub.log.debug(
                            f"Requisite {req_type} {state}:{name} not found in current run. checking in ESM for requisite"
                        )
                        r_chunks_from_esm = hub.tool.idem.esm.get_chunks_from_esm(
                            state, name
                        )
                        if not r_chunks_from_esm:
                            hub.log.debug(
                                f"Requisite {req_type} {state}:{name} not found in ESM."
                            )
                            data["errors"].append(
                                f"Requisite {req_type} {state}:{name} not found in ESM."
                            )
                        for r_chunk in r_chunks_from_esm:
                            # checking if this chunk is sls.run, sls.run is a structure
                            # that runs a group of files, we need to find the exact
                            # chunk in that group of files based on args and set that chunk in reqrets.
                            if hub.tool.idem.sls.is_state_has_sls_run(r_chunk):
                                (
                                    sls_run_chunk,
                                    new_args,
                                ) = hub.idem.tools.get_sls_run_chunks(
                                    low, state, name, args, from_esm=True
                                )
                                if sls_run_chunk:
                                    r_chunk = sls_run_chunk
                                    args = new_args

                            reqret_esm = hub.tool.idem.esm.update_running_from_esm(
                                r_chunk
                            )
                            r_tag = hub.idem.tools.gen_chunk_func_tag(r_chunk)
                            reqret = {
                                "req": req_type,
                                "name": name,
                                "state": state,
                                "r_tag": r_tag,
                                "ret": reqret_esm,
                                "chunk": "r_chunk",
                                "args": args,
                            }
                            data["reqrets"].append(reqret)
    return seq



from typing import Any
from typing import Dict


async def put(
    hub,
    body: Any,
    profile: str = "default",
    tags: Dict[str, Any] = None,
):
    if tags is None:
        tags = {}

    await hub.evbus.broker.put(
        profile=profile, body=dict(tags=tags, message=body, run_name=hub.idem.RUN_NAME)
    )


def put_nowait(
    hub,
    body: Any,
    profile: str = "default",
    tags: Dict[str, Any] = None,
):
    if tags is None:
        tags = {}

    hub.evbus.broker.put_nowait(
        profile=profile, body=dict(tags=tags, message=body, run_name=hub.idem.RUN_NAME)
    )



def sig_is_pending(hub, ret: dict, state: str = None):
    """
    Validate the signature of the pending plugin function
    """
    ...


def post_is_pending(hub, ctx):
    """
    Conform the return structure of is_pending is a boolean
    """
    assert isinstance(ctx.ret, bool), ctx.ret
    return ctx.ret



from typing import List

SLS_RUN_SEPARATOR = ":sls:"

__IS_SLS_RUN__ = True

REQUISITES_TO_UPDATE_FOR_SLS_RUN = ["require", "require_in", "require_any", "listen"]


async def run(
    hub,
    ctx,
    name: str,
    sls_sources: List[str],
    params: List[str] = None,
    **kwargs,
):
    """
    State module to structure a group of SLS files into an independent construct that
    can be invoked from SLS code multiple times with different set of arguments and parameters

    Args:
        name(string):
            An idem name of the resource.

        sls_sources(list):
            List of sls files to run

        params(list, Optional):
            List of params files. All the params provided in these files will be consolidated and these consolidated params
            will override the params passed in Idem run. These consolidated params along with params passed to Idem run
            will be used to run the sls files provided in sls_sources. Defaults to None.

        kwargs(Dict[str, Any], Optional):
            parameters passed in kwargs are used as parameters to resolve parameters in sls_sources files.

        Request Syntax:
           .. code-block:: sls

              [sls-run-name]:
                sls.run:
                  - sls_sources:
                      - 'string'
                  - params:
                      - 'string'

        Returns:
            Dict[str, Any]

        Examples:
           .. code-block:: sls

               service4:
                sls.run:
                  - sls_sources:
                      - sls.service_file1
                      - sls.service_file2
                  - params:
                      - params.file1
    """

    result = dict(comment=[], name=name, old_state=None, new_state=None, result=True)

    # Name of the sls.run state if this state is inside another sls.run.
    # This parameter is used by idem execution internally to pass
    # current sls_run_id to the nested sls.run state executions
    sls_run_id = kwargs.get("sls_run_id", None)

    # The sls_sources and params provided to this state are paths to files.
    # To resolve the location of files and parse them we use this hub.OPT.idem.tree which gives us the base path.
    sls_sources_path = list()
    param_sources_path = list()
    if hub.OPT.idem.tree:
        tree = f"file://{hub.OPT.idem.tree}"
        hub.log.debug(f"Added tree to sls and param sources: {tree}")
        sls_sources_path.append(tree)
        param_sources_path.append(tree)

    main_run_name = ctx.run_name
    # Create a temporary run name to use while compiling the sls_sources passed to this state
    temporary_run_name = main_run_name + "." + name

    # Create a temporary run to process the new sls block
    await hub.idem.state.create(
        name=temporary_run_name,
        sls_sources=sls_sources_path,
        # Allow a different render pipe to be used for the new render block
        # Default to the renderer of the main run
        render=hub.idem.RUNS[main_run_name].get("render"),
        # Copy state.apply parameters from the main run
        **{
            k: hub.idem.RUNS[main_run_name].get(k)
            for k in (
                "runtime",
                "subs",
                "cache_dir",
                "test",
                "acct_file",
                "acct_key",
                "acct_profile",
                "acct_blob",
                "managed_state",
                "param_sources",
                "acct_data",
                "invert_state",
            )
        },
    )

    # Gather params files provided to this file and combine with idem run params
    # we gather params from idem run, params files provided to this state and inline params provided to this state.
    # will combine the params from all the three sources. If there are common params, Inline params takes precedence
    # followed by params files provided to this state.
    run_params_ret = await _gather_params(
        hub, main_run_name, temporary_run_name, params, param_sources_path, kwargs
    )
    if "errors" in run_params_ret:
        result["comment"] = [
            f"Error in gathering params files for sls.run {name}"
        ] + run_params_ret["errors"]
        result["result"] = False
        hub.idem.RUNS.pop(temporary_run_name, None)
        return result

    # parse the sls_sources provided to this state with the consolidated params and get high data
    src = hub.idem.sls_source.init.get_refs(sources=sls_sources_path, refs=sls_sources)

    # Resolve the new sls_sources with the main run
    gather_data = await hub.idem.resolve.init.gather(
        temporary_run_name,
        *src["sls"],
        sources=src["sls_sources"],
    )
    # Add the newly resolved blocks to the temporary run
    await hub.idem.sls_source.init.update(temporary_run_name, gather_data)

    if hub.idem.RUNS[temporary_run_name]["errors"]:
        result["comment"] = [
            f"Error in gathering sls_sources for sls.run {name}"
        ] + hub.idem.RUNS[temporary_run_name]["errors"]
        result["result"] = False
        hub.idem.RUNS.pop(temporary_run_name, None)
        return result

    # update the requisites in the high data of sls_sources passed to this state to include sls_run_id
    _update_requisites(hub, name, temporary_run_name, sls_run_id)

    # compile the high data to low data
    await hub.idem.state.compile(temporary_run_name)

    low_data = hub.idem.RUNS[temporary_run_name]["low"]

    sls_run_id = f"{sls_run_id}{SLS_RUN_SEPARATOR}{name}" if sls_run_id else name
    # Iterate over the states passed to this sls.run and add the extra attributes __sls_run_idm, sls_run_id
    # to identify these states from other states that are running in main idem run.
    for chunk in low_data:
        # sls_run_id is added to chunk to make requisite system understand this chunk is a part of sls.run
        chunk["sls_run_id"] = sls_run_id

        # sls_run_id is appended to __id__ of chunk to make this chunk unique as
        # different sls.run states can include same files
        chunk["__id__"] = f"{sls_run_id}:{chunk['__id__']}"

        # Add the low data gathered from sls_sources passed to this state to main run
        hub.idem.RUNS[main_run_name].get("add_low").append(chunk)

    result["new_state"] = {
        "Status": f"Added {len(low_data)} states to be run.",
        "is_sls_run": True,
    }
    hub.idem.RUNS.pop(temporary_run_name, None)
    return result


def is_pending(hub, ret):
    """
    Always skip reconciliation for sls.run because is_pending will be called for individual states
    """
    return False


# Utility methods to extract params and sls_sources passed to sls.run and update requisites


async def _gather_params(
    hub, main_run_name, temporary_run_name, params, param_sources_path, kwargs
):
    """Gather params passed to main Idem run, params files passed to the sls.run
       and inline parameters passed to sls.run
       will combine the params from all the three sources. If there are common params, Inline params takes precedence
       followed by params files provided to this state.

    Args:
        main_run_name(string):
            Name of the Idem run

        params(Dict[str, List[str]]):
            params sources and params files to get params from.

        param_sources_path(list):
            List of file paths of provided params

        kwargs(Dict[str, Any]):
            Inline params provided to the sls.run
    """
    run_params = {}
    # params given during Idem main run.
    run_params.update(hub.idem.RUNS[main_run_name]["params"] or {})

    # parse params files provide to this state.
    if params:
        params = hub.idem.sls_source.param.get_refs(
            sources=param_sources_path, refs=params
        )
        resolved_params_ret = await _gather_params_from_included_files(
            hub, temporary_run_name, params
        )
        if not resolved_params_ret or "errors" in resolved_params_ret:
            return {"errors": resolved_params_ret["errors"]}
        # combining the params provided to this file and idem run params.
        # if there is overlapping of params, params provided to this file will take precedence over idem run params.
        if "params" in resolved_params_ret:
            run_params.update(resolved_params_ret["params"])

    # update the params with inline params if provided.
    if kwargs:
        run_params.update(kwargs)
    hub.idem.RUNS[temporary_run_name]["params"] = run_params
    return {"params": run_params}


async def _gather_params_from_included_files(hub, temporary_run_name, params):
    """Gather parameters from the params files.

    Args:
        main_run_name(string):
            Name of the Idem run

        params(Dict[str, List[str]]):
            params sources and params files to get params from.
    """

    param_sources = params["param_sources"]
    params = params["params"]
    gather_data = await hub.idem.resolve.init.gather(
        temporary_run_name, *params, sources=param_sources
    )
    if gather_data["errors"]:
        return {"errors": gather_data["errors"]}
    hub.idem.sls_source.param.process_params(temporary_run_name, gather_data)

    return {"params": hub.idem.RUNS[temporary_run_name]["params"]}


def _update_requisites(hub, name, run_name, sls_run_id=None):
    _format_argument_binding(hub, name, run_name, sls_run_id)
    _format_require(hub, name, run_name, sls_run_id)


def _format_argument_binding(hub, name, run_name, sls_run_id=None):
    """Update argument binding requisite in the high data with the sls.run name
    Args:
        name(string):
            An Idem name of sls.run state

        run_name(string):
            Name of current sls.run combined with idem run name. <idem_run_name>.<sls_run_id>

        sls_run_id(string, Optional):
            Name of the sls.run state that this state is part of. If this sls.run is nested inside another sls.run
            sls_run_id is name of the parent sls.run state.

    """

    high_data = hub.idem.RUNS[run_name]["high"]
    # Iterate through high data leaf arguments
    for id_, body, state, run, arg_key, arg_value in hub.idem.tools.iter_high_leaf_args(
        high_data
    ):
        if (
            arg_value
            and isinstance(arg_value, str)
            and hub.idem.ccomps.arg_bind.ARG_REFERENCE_REGEX_PATTERN.search(arg_value)
        ):
            # checking for argument binding syntax
            for arg_ref in hub.idem.ccomps.arg_bind.ARG_REFERENCE_REGEX_PATTERN.findall(
                arg_value
            ):
                state_name = arg_ref.split(":")[1]
                # check the referred state is present in this sls.run
                # If it's present, update the argument binding by adding sls:<sls.run_name> prefix
                # This ensures we are referring correct state when we add this data to Idem run
                if state_name in high_data:
                    updated_arg_binding = arg_ref.replace(
                        "${",
                        f"${{sls:{(sls_run_id + SLS_RUN_SEPARATOR) if sls_run_id else ''}{name}:",
                    )
                    _replace_arg_bind_in_high_data(
                        hub,
                        high_data.get(id_).get(state),
                        arg_key,
                        arg_ref,
                        updated_arg_binding,
                    )
    hub.idem.RUNS[run_name]["high"] = high_data


def _format_require(hub, name, run_name, sls_run_id=None):
    """Update require requisite in the high data with the sls.run name
    Args:
        name(string):
            An Idem name of sls.run state

        run_name(string):
            Name of current sls.run combined with idem run name. <idem_run_name>.<sls_run_id>

        sls_run_id(string, Optional):
            Name of the sls.run state that this state is part of. If this sls.run is nested inside another sls.run
            sls_run_id is name of the parent sls.run state.

    """
    high_data = hub.idem.RUNS[run_name]["high"]
    for id_, body in high_data.items():
        if id_.startswith("__"):
            continue
        for state in body:
            if state.startswith("__"):
                continue
            for arg in body[state]:
                # update require requisite key to include sls run name.
                if isinstance(arg, dict):
                    for attribute_key, attribute_val in arg.items():
                        if attribute_key in REQUISITES_TO_UPDATE_FOR_SLS_RUN:
                            updated_require_requisites = []
                            for require_state in attribute_val:
                                require_arg_key = next(iter(require_state))
                                require_arg_value = require_state[require_arg_key]
                                updated_require_requisites.append(
                                    _prepare_sls_run_require_requisite(
                                        name,
                                        require_arg_key,
                                        require_arg_value,
                                        sls_run_id,
                                    )
                                )
                            arg[attribute_key] = updated_require_requisites
    hub.idem.RUNS[run_name]["high"] = high_data


def _replace_arg_bind_in_high_data(
    hub, high_data_resource, arg_key, arg_value, arg_value_to_replace
):
    """Replace the value in high data"""
    key_arr = arg_key.split(":")
    referred_key = key_arr.pop(0)
    key_to_parse, indexes = hub.tool.idem.arg_bind_utils.parse_index(referred_key)
    for index, resource_state in enumerate(high_data_resource):
        name_def = next(iter(resource_state))
        if key_to_parse == name_def:
            hub.idem.rules.arg_resolver.set_chunk_arg_value(
                resource_state,
                arg_value,
                arg_key.split(":"),
                arg_value_to_replace,
                None,
            )


def _prepare_sls_run_require_requisite(
    name, require_arg_key, require_arg_value, sls_run_id
):
    """
    modify require requisite to include sls.run name so that we point out the exact required resource
    This modification ensures that we require the resource in this sls.run only

    Example-:
        In the resources inside this sls.run, user will add require requisite like this

        state1:
         cloud1.state_def:
          - require:
             - cloud2: state2

        This same state "state1" and "state2" might be part of more than one sls.run. so to identify which exact state
        to require we need to add the sls.run name.

        Modified state:

            state1:
             cloud1.state_def:
              - require:
                 - sls: <sls.run.name>
                    - cloud2: state2

        If we have nested sls.runs, then sls_run_id is passed.
        Example-:

        state3 is a sls.run which contains state4 and state3 is part of State1 sls.run

        state4:
          test.present:
            - require:
                - test: test_include_sls_run_3

        This will be modified as

        state4:
          test.present:
            - require:
                - sls: state1
                    - sls: state3
                        - test: test_include_sls_run_3


    """
    require_sls_requisite = {}
    current_obj = [{require_arg_key: require_arg_value}]
    if sls_run_id:
        # If it's a nested sls.run we split the sls_run_id and add require requisites in hierarchy
        sls_run_states = sls_run_id.split(SLS_RUN_SEPARATOR)
        # Adding the current or innermost sls.run name
        sls_run_states.append(name)
        # loop through in reverse order since inner sls.run should be added first
        for sls_run_state in reversed(sls_run_states):
            current_obj = [{"sls": [{sls_run_state: current_obj}]}]
        # we need to output a dictionary so sending the first element of array
        # as array will always contain only one element
        return current_obj[0]
    else:
        require_sls_requisite["sls"] = [{name: current_obj}]
        return require_sls_requisite



from pathlib import Path
from typing import Dict

from jinja2 import Template


async def render(
    hub,
    ctx,
    name: str,
    template_file_name: str = None,
    template: str = None,
    variables: Dict = None,
):
    """
    A state to render data from jinja template file or an embedded jinja template.
    Templates are a useful way to break off complex logic and have shared functionality.

    Args:
        name(Text): The name of the state.
        template_file_name(Text, Optional): Full path to a jinja template file.
        template(Text, Optional): Embedded Jinja template within {% raw %} {% endraw %}.
            either template or template_file_name should be specified
        variables(Dict, Optional): Variables for interpolation within the template.

    Returns:
        {"result": True|False, "comment": "A message", "old_state": {},
        "new_state": {"rendered": output_content}}

    Request Syntax:
        [output-id]:
          template.render:
           - name: 'string'
           - template_file_name: 'string'
           - variables: Dict

    Example: aws_auth_config_map.tpl has the below content which can be shared in multiple sls files.

    .. code-block::

        {% for user in users_list %}
        - userarn: arn:aws:iam::{{ aws_caller_identity }}:user/{{ user }}
          username: {{ user }}
          groups:
           - system:masters
        {% endfor %}
        {% for svc_user in svc_user %}
        - userarn: arn:aws:iam::{{ aws_caller_identity }}:user/eks/{{ svc_user }}/{{ svc_user }}
          username: {{ svc_user }}
          groups:
           - cluster: {{ svc_user }}
        {% endfor %}

    .. code-block:: sls

        template_render_test:
          template.render:
            - name: template_render_test
            - template_file_name: "/home/files/templates/aws_auth_config_map.tpl"
            - variables:
                users_list: ["user1", "user2"]
                svc_user: ["user1", "user2"]
                aws_caller_identity: "123456789012"

        kubernetes_config_map.aws_auth:
          k8s.core.v1.config_map.present:
            - metadata:
                name: "aws-auth"
                namespace: "kube-system"
            - data:
                mapUsers: ${template:template_render_test:rendered}


        After rendering the template 'aws_auth_config_map.tpl', the output of the state 'template_render_test' will be:

        - userarn: arn:aws:iam::123456789012:user/user1
          username: user1
          groups:
            - system:masters

        - userarn: arn:aws:iam::123456789012:user/user2
          username: user2
          groups:
            - system:masters

        - userarn: arn:aws:iam::123456789012:user/eks/user1/user1
          username: user1
          groups:
            - cluster: user1

        - userarn: arn:aws:iam::123456789012:user/eks/user2/user2
          username: user2
          groups:
            - cluster: user2

        this rendered data will be passed as input to the state 'kubernetes_config_map.aws_auth'

    """
    result = dict(
        comment=(), old_state={}, new_state={}, changes={}, name=name, result=True
    )
    output_content = ""

    if template_file_name is None and template is None:
        result["comment"] = (
            "Either template_file_name or template should be provided",
        )
        result["result"] = False
        return result

    # Jinja template file
    if template_file_name is not None:
        template_file = Path(template_file_name)
        if template_file.is_file() is False:
            msg = f"Template file not found: {template_file_name}"
            hub.log.debug(msg)
            result["result"] = False
            result["comment"] = (msg,)
            return result
        else:
            with open(template_file) as curr_template_file:
                tm = Template(curr_template_file.read())
                output_content = tm.render(**variables) if variables else tm.render()
    # Embedded Jinja template
    elif template is not None:
        tm = Template(template)
        output_content = tm.render(**variables) if variables else tm.render()

    result["comment"] = (f"Template rendering is success",)
    # The final rendered template
    result["new_state"] = {"rendered": output_content}

    # to display the rendered data in output, we need to set it in result["changes"], as changes
    # will not be derived for this state like other resource states.
    result["changes"] = {"rendered": output_content}
    return result


def is_pending(hub, ret):
    print("executing is pending for template")
    return not ret["result"] is True



import asyncio
from typing import Dict


async def sleep(hub, ctx, name, duration: str = None) -> Dict[str, Dict]:
    """
    Sleep resource to be used by other resources to delay creation/deletion for a duration of time in seconds

    Examples:
        .. code-block:: sls

            sleep_60s:
                time.sleep:
                - duration: 60

            some_machine:
                cloud.instance.present:
                - name: my-instance
                - require:
                  - time.sleep: sleep_60s

    """
    result = dict(
        comment=(),
        old_state={},
        new_state={},
        name=name,
        result=True,
    )

    if duration is None or str(duration) == 0:
        result["result"] = False
        result["comment"] = ("Duration is required.",)
        return result

    if ctx.get("test", False):
        result["comment"] = (f"Would sleep for {duration} seconds.",)
        return result

    await asyncio.sleep(duration)

    result["comment"] = (f"Successfully slept for {duration} seconds.",)

    return result



import copy
from typing import Dict


async def runtime(
    hub,
    name: str,
    ctx: Dict,
    seq: Dict,
    low: Dict,
    running: Dict,
    managed_state: Dict,
    progress,
):
    """
    Execute the runtime in parallel mode
    """
    inds = []
    for ind in seq:
        if seq[ind].get("unmet"):
            # Requisites are unmet, skip this one
            continue
        inds.append(ind)
    if not inds:
        # Nothing can be run, we have hit recursive requisite,
        # or we are done
        pass

    for ind in inds:
        await hub.idem.rules.init.run(
            name,
            # every state run should use unique ctx so that ctx.acct retains correct information during whole state execution.
            copy.deepcopy(ctx),
            low,
            seq[ind],
            running,
            hub.idem.RUNS[name]["run_num"],
            managed_state,
            seq,  # need this object in recreate_on_update rule check, to find the dependents for current seq comp.
        )
        hub.idem.RUNS[name]["run_num"] += 1
        hub.tool.progress.init.update(progress)



from typing import Any
from typing import Dict


async def modify(hub, name: str, chunk: Dict[str, Any]) -> Dict[str, Any]:
    """
    Take the given run name and low chunk, and allow state plugins to modify
    the low chunk
    """
    for mod in hub.idem.mod:
        if mod.__name__ == "init":
            continue
        if hasattr(mod, "modify"):
            chunk = mod.modify(name, chunk)
            chunk = await hub.pop.loop.unwrap(chunk)
    return chunk



from typing import Any
from typing import Dict


def apply(hub, data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Reorganize the data by sorting by each state's unique tag
    """
    result = {}
    for tag in sorted(data.keys()):
        result[tag] = data[tag]
    return result



"""
These classes are used to extract paramaters used inside
SLS files during validation. The code works by simulating
get() function of Python Dictionary object. As we get any
request to access parameter foo using {{ params.get('foo') }}
in SLS file we capture that request sending back a dummy
value, and identifying the parameter name in the process.

To give specific example, a SLS file like this:

    Assure Resource Group Present {{ params.get('rg_name').get('your_rg') }}:
    azure.resource_management.resource_groups.present:
        - resource_group_name: {{ params.get('rg_name').get('your_rg') }}
        - parameters:
            - location{{ params['locations'][4].name }}

    Assure Resource Group Present {{ params.get('rg_name').get('my_rg', 'rg') }}:
    azure.resource_management.resource_groups.present:
        - resource_group_name: {{ params.get('rg_name').get('my_rg', 'rg') }}
        - parameters:
            - location{{ params['locations'][0].name }}

will produce parameter section like the following in
validate sub command output:

    "parameters": {
        "rg_name": {
            "your_rg": "",
            "my_rg": "rg"
        },
        "locations": [
            {
                "name": ""
            }
        ]
    }

TODO: Implement rest of Python dict functionality in Parameters

"""


class BaseParam:
    def __init__(
        self, parent, name, display, is_indexing_operator, internal, default
    ) -> None:
        self._parent = parent
        self._name = str(name)
        self._display = BaseParam._get_display(
            name, display, is_indexing_operator, default
        )
        self._internal = internal
        self._default = "" if default is None else str(default)
        self._params = {}
        self._is_leaf = False
        self._attributes = set()

    @staticmethod
    def _get_display(name, parent_display, is_indexing_operator, default):
        if parent_display is None:
            return ""
        ret = "params" if len(parent_display) == 0 else f"{parent_display}"
        if isinstance(name, int):
            return f"{ret}[{name}]"
        elif is_indexing_operator:
            return f"{ret}['{name}']"
        else:
            return (
                f"{ret}.get('{name}')"
                if default is None
                else f"{ret}.get('{name}', '{default}')"
            )

    def params(self):
        if not self._params:
            return self._default

        ret = None
        for key, value in self._params.items():
            if isinstance(key, int):
                if ret is None:
                    ret = []
                for i in range(len(ret), key + 1):
                    ret.append(None)
                ret[key] = value.params()
            else:
                if ret is None:
                    ret = {}
                ret[key] = value.params()

        return ret

    def _create_child(self, param, is_indexing_operator, default=None):
        if self._is_leaf:
            raise TypeError(
                f"Terminal leaf can not have atrributes: {self._display} can not have attribute {param}."
            )

        if isinstance(param, int):
            internal = f"{self._internal}[{param}]"
            child = ParamArray(
                self, param, self._display, is_indexing_operator, internal, default
            )
        else:
            internal = (
                f"^^{param}^^"
                if self._internal == "" or self._internal is None
                else f"{self._internal}.~~{param}"
            )
            child = ParamObject(
                self, param, self._display, is_indexing_operator, internal, default
            )

        self._params[param] = child
        return child

    def _get_value(self, param):
        if param in self._params:
            return self._params[param]

    def _set_display(self, display):
        self._display = display

    def __getitem__(self, param):
        return self.get(param)

    def __str__(self) -> str:
        return f"?? {self._display} ?? {self._internal} ??"

    def items(self):
        return ResponseObject(self, tuple, 2)

    def __getattr__(self, name):
        if name not in self._attributes:
            self._attributes.add(name)
        raise AttributeError(
            f"Function {self._name} is not supported on {self._parent._display}"
        )

    def __call__(self, *args, **kwargs):
        return self


class ResponseObject:
    def __init__(self, item, ret_type=tuple, pack_count=0):
        self._item = item
        self._next_index = 0
        self._index = 0
        self._ret_type = ret_type
        self._pack_count = pack_count

    def __len__(self):
        return self._pack_count

    def __getitem__(self, index):
        if self._index == 0:
            self._index += 1
            if self._pack_count == 0:
                return self._item

            t = []
            for i in range(self._pack_count):
                t.append(str(self._item))
            return self._ret_type(t)

        self._index = 0
        raise StopIteration

    def __call__(self, *args, **kwargs):
        return self

    def __next__(self):
        if self._next_index == 0:
            self._next_index += 1
            return self
        self._next_index = 0
        raise StopIteration

    def __str__(self):
        return str(self._item)


class ParamObject(BaseParam):
    def __init__(
        self, parent, name, display, is_indexing_operator, internal, default
    ) -> None:
        super().__init__(parent, name, display, is_indexing_operator, internal, default)

    def _get_common(self, param, is_indexing_operator, default):
        child = super()._get_value(param)
        if child is not None:
            return child

        if param in self._attributes:
            return ResponseObject(self)

        return super()._create_child(param, is_indexing_operator, default)

    def get(self, param, default=None):
        return self._get_common(param, False, default)

    def __getitem__(self, param):
        return self._get_common(param, True, None)

    def __iter__(self):
        return ResponseObject(self, ResponseObject, 2)


class ParamArray(ParamObject):
    def __init__(
        self, parent, name, display, is_indexing_operator, internal, default
    ) -> None:
        super().__init__(parent, name, display, is_indexing_operator, internal, default)

    def __str__(self) -> str:
        self._is_leaf = True
        return super().__str__()


class Parameters(ParamArray):
    def __init__(self) -> None:
        super().__init__(None, None, None, False, None, None)


def get_validate_params(hub):
    return Parameters()



import uuid
from collections import ChainMap
from typing import List

import dict_tools.differ as differ


async def build(hub, ctx, name: str, triggers: List = None, resource_id: str = None):
    """
    A state to represent a trigger. A trigger will be a dictionary of key-values,
    which when changed, can be used to trigger another state. The dictionary can
    hold values that are user defined e.g. last_run_id: {{ range(1, 51) | random }}
    or it can hold values from the output of a different state as supported in arg
    binding e.g. user_name: ${aws.iam.user:aws_iam_user.extension-jenkins:user_name}.

    Note: Triggers dictionary can hold values which won't change over the course
    of multiple execution. The example below is a constantly changing trigger state.


    :param hub:
    :param ctx:
    :param name: The name of the state
    :param resource_id: The resource_id to represent this trigger in ESM. Its an internal id to track
                 whether this resource has run or not.
    :param triggers: The dictionary to hold key-value pairs

    .. code-block:: sls

        always-changes-and-succeeds:
          test.succeed_with_changes:
            - name: foo

        always-changes-trigger:
            trigger.build:
                - triggers:
                     - last_run_id: {{ range(1, 51) | random }}
                     - comment: ${test:always-changes-and-succeeds:testing}

        watch_changes:
          test.nop:
            - onchanges:
              - trigger: always-changes-trigger

    In the example above , Only when there are `changes` when the aws.trigger.build is run , the `local-command` state runs.
    """
    result = dict(comment=(), old_state={}, new_state={}, name=name, result=True)
    before = None

    if resource_id:
        before = ctx.get("old_state")
    else:
        resource_id = "idem-trigger-" + str(uuid.uuid4())
    if before:
        result["old_state"] = before

    if triggers:
        triggers = dict(ChainMap(*triggers))

    result["new_state"] = dict(triggers=triggers, resource_id=resource_id)

    try:
        result["changes"] = differ.deep_diff(
            result["old_state"].get("triggers", {}),
            result["new_state"].get("triggers", {}),
        )

        if result["changes"]:
            result["comment"] = (f"Trigger '{name}' built with changes",)
        else:
            result["comment"] = (f"Trigger '{name}' built with no changes",)
    except Exception as e:
        result["result"] = False
        result["comment"] = result["comment"] + (str(e),)

    return result



from typing import Dict
from typing import List

import dict_tools.data as data


def __init__(hub):
    # Keep track of the number of times the idem runtime is called for the progress bar
    hub.idem.run.init.RUNTIME_ITERATIONS = 0


async def start(hub, name, pending_tags: List = None, target: str = None):
    """
    Called only after the named run has compiled low data. If no low data
    is present an exception will be raised
    pending_tags: list of state tags to re-run, passed from reconciliation.
                  If None all states are executed.
    """
    if not hub.idem.RUNS[name].get("low"):
        raise ValueError(f"No low data for '{name}' in RUNS")
    ctx = data.NamespaceDict({"run_name": name, "test": hub.idem.RUNS[name]["test"]})

    rtime = hub.idem.RUNS[name]["runtime"]
    low = hub.idem.RUNS[name].get("low")
    # Add metadata to second run

    # Fire an event with all the pre-rendered low-data for all states
    await hub.idem.event.put(
        profile="idem-low",
        body=low,
        tags={"ref": "idem.run.init.start", "type": "state-low-data"},
    )
    old_seq = {}
    old_seq_len = -1
    needs_post_low = True
    options = {"invert_state": hub.idem.RUNS[name]["invert_state"]}
    low_declaration_ids = [item.get("__id__") for item in low]

    if pending_tags:
        # During reconciliation only pending tags and their requisites that
        # are not other resources - such as sleep/script will be re-run.
        # These resources will be recursively evaluated and added to the low data.

        # Getting rerun_data from all tags
        rerun_data_map = {}
        for tag in hub.idem.RUNS[name]["running"]:
            if hub.idem.RUNS[name]["running"][tag].get("rerun_data"):
                rerun_data_map[tag.split("_|-")[1]] = hub.idem.RUNS[name]["running"][
                    tag
                ].get("rerun_data")
        # Clearing previous running data.
        hub.idem.RUNS[name]["running"] = {}
        declaration_ids = [tag.split("_|-")[1] for tag in pending_tags]
        low = extend_pending_tags(low, declaration_ids, rerun_data_map)

    if target:
        # If target is specified - identify all its dependencies
        low_items = gather_target_and_reqs(low, target)
        if len(low_items) < 1:
            raise ValueError(
                f"Invalid 'target' for run '{name}': {target}. 'target' should be a declaration ID."
            )

        declaration_ids = [item.get("__id__") for item in low_items]
        hub.idem.RUNS[name]["high"] = {
            key: item
            for (key, item) in hub.idem.RUNS[name]["high"].items()
            if key in declaration_ids
        }
        low = low_items

    while True:
        seq = hub.idem.req.seq.init.run(
            None, low, hub.idem.RUNS[name]["running"], options
        )
        if seq == old_seq:
            unmet_reqs = {chunk["tag"]: chunk["unmet"] for chunk in seq.values()}
            if len(unmet_reqs) == 0:
                raise Exception(f"Invalid syntax for '{name}'")
            raise Exception(f"No sequence changed for '{name}': {unmet_reqs}")

        # Initialize the progress bar
        progress_bar = hub.tool.progress.init.create(
            seq,
            desc=f"idem runtime: {hub.idem.run.init.RUNTIME_ITERATIONS}",
            unit="states",
        )
        # Increment the runtime iteration number after using it for the progress bar
        hub.idem.run.init.RUNTIME_ITERATIONS += 1

        await hub.idem.run[rtime].runtime(
            name,
            ctx,
            seq,
            low,
            hub.idem.RUNS[name]["running"],
            hub.idem.RUNS[name]["managed_state"],
            progress=progress_bar,
        )

        render_data = await hub.idem.resolve.init.render(
            name,
            blocks=hub.idem.RUNS[name]["blocks"],
            sls_refs=hub.idem.RUNS[name]["sls_refs"],
            resolved=hub.idem.RUNS[name]["resolved"],
        )
        await hub.idem.sls_source.init.update(name, render_data)

        await hub.idem.state.compile(name)

        extend_low(hub, name, low, low_declaration_ids)

        await hub.idem.event.put(
            profile="idem-low",
            body=low,
            tags={"ref": "idem.run.init.start", "type": "state-low-data"},
        )
        if len(low) <= len(hub.idem.RUNS[name]["running"]):
            if hub.idem.RUNS[name]["post_low"] and needs_post_low:
                hub.idem.RUNS[name]["low"].extend(hub.idem.RUNS[name]["post_low"])
                needs_post_low = False
                extend_low(hub, name, low, low_declaration_ids)
                continue
            else:
                break
        if len(seq) == old_seq_len:
            # check if the sequence chunks are same or not.
            # if same then throw RecursionError else proceed with the flow.
            # this is because, when we add chunk to delete old_resource as part of recreate_on_update requisite
            # length of old_seq and seq are same, but with different sequence components.
            # so we need to compare the chunks of seq and old_seq.

            # generate the tags for old chunks
            old_tags = {
                hub.idem.tools.gen_chunk_func_tag(old_seq.get(index).get("chunk"))
                for index in old_seq
            }

            # generate the tags for current chunks
            current_tags = {
                hub.idem.tools.gen_chunk_func_tag(seq.get(index).get("chunk"))
                for index in seq
            }

            # if old_tags and current_tags are same, then throw RecursionError
            if len(old_tags - current_tags) == 0:
                raise RecursionError(
                    f"No progress made on '{name}', Recursive Requisite!"
                )
        old_seq = seq
        old_seq_len = len(seq)


def gather_target_and_reqs(low, target) -> List[dict]:
    """
    Gather the target state(s) and all its requisites
    :param low: low data
    :param target: target name: should match declaration id
    :return: list of chunks
    """
    low_items = []

    # Target must be declaration id
    # It might resolve to multiple target items if there are
    # multiple resources under the same declaration id
    gather_(low, [target], low_items, None)

    return low_items


def gather_(low: List[Dict], reqs: List[str], low_items: List[Dict], filters: list):
    """
    Recursively gather low items of target(s) and their pre-requisites.
    Any reqs - declaration ID might be resolved to multiple resources,
    in case there are multiple resources under the same declaration id.
    :param low:  low data
    :param reqs: list of declaration ids of requisites ('require')
    :param low_items: gathered required low items
    :param filters: filters to filter out low items
    :return: low data to re-run
    """
    if not reqs:
        return
    for req in reqs:
        req_items = [item for item in low if item.get("__id__") == req]
        # Start filtering only after first iteration, where original item(s) are added
        if low_items and filters:
            req_items = filter_out_(req_items, filters)
        if len(req_items) < 1:
            continue
        low_items.extend(req_items)
        for req_item in req_items:
            if req_item.get("require"):
                for required in req_item.get("require"):
                    gather_(low, list(required.values()), low_items, filters)


def extend_low(hub, name, low, low_declaration_ids):
    # If new chunks were added to "low" (post_low etc)
    # Add them in case low was modified by 'pending_tags' or 'target'
    new_low_items = [
        item
        for item in hub.idem.RUNS[name].get("low")
        if item.get("__id__") not in low_declaration_ids
    ]
    low.extend(new_low_items)
    low_declaration_ids += [item.get("__id__") for item in new_low_items]


def extend_pending_tags(
    low, pending_decl_id: list, rerun_data_map: dict = None
) -> List[dict]:
    """
    Gather the states that are corresponding to the pending tags,
    and its 'require' dependencies that are not resources,
    Such as time.sleep or data.write or exec.
    :param low: low data
    :param pending_decl_id: list of pending declaration ids from reconciliation
    :return: low items for pending states and their dependencies
    """
    low_items = []

    # Pending declaration ids might resolve to multiple items if there are
    # multiple resources under the same declaration id
    gather_(low, pending_decl_id, low_items, [{"fun": "absent"}, {"fun": "present"}])

    for key, rerun_data in rerun_data_map.items():
        for item in low_items:
            if item.get("__id__") == key:
                item["rerun_data"] = rerun_data

    return low_items


def filter_out_(low_items: List[Dict], filters: List[Dict]) -> List[Dict]:
    # Filter out required low items that are resources (present/absent)
    # as those will be retrieved from the ESM. But any others exec/sleep/script
    # required will be executed
    if not low_items or not filters:
        return low_items

    new_low_items = []
    for item in low_items:
        match = False
        for filter in filters:
            for key, value in filter.items():
                if item.get(key) == value:
                    match = True
        if not match:
            new_low_items.append(item)

    return new_low_items



def is_pending(hub, ret: dict, state: str = None) -> bool:
    """
    Default implementation of pending plugin
    Returns 'True' when the state is still 'pending' and reconciliation is required.

    If the state implements is_pending() call the state's implementation,
    otherwise state is pending until 'result' is 'True' and there are no 'changes'
    for normal flows, state is pending until 'result' is 'True' for resource recreation flow.

    :param hub: The hub
    :param ret: (dict) Returned structure of a run
    :param state: (Text, Optional) The name of the state
    :return: bool
    """

    if (
        state is not None
        and hub.states[state] is not None
        and callable(getattr(hub.states[state], "is_pending", None))
    ):
        return hub.states[state].is_pending(ret=ret)

    # for resource recreation flow, we should consider only whether the result is True or not.
    if ret.get("recreation_flow", False):
        return not ret["result"] is True

    return not ret["result"] is True or bool(ret["changes"])



from typing import Any
from typing import Dict
from typing import List
from typing import Tuple


SKIP_ESM = True


async def run(
    hub,
    ctx,
    name: str,
    *,
    path: str = None,
    acct_profile: str,
    acct_data=None,
    args: List[Any] = None,
    kwargs: Dict[str, Any] = None,
):
    """
    Call an exec module by path reference

    Args:
        hub:
        ctx:
        name(Text): The name of the state or the exec module reference path
        path(Text): The exec module reference path to call
        acct_profile(Text): The acct profile to use on the exec module call
        acct_data(Dict): The acct_data to use with the exec module call
        args(List): A list of args to pass to the exec module
        kwargs(Dict): The keyword arguments to pass to the exec module

    Returns:
        {"result": True|False, "comment": ["A message"], "new_state": The return from the exec module}


    .. code-block:: yaml

        exec_func:
          exec.run:
            - path: test.ping
            - acct_profile: default
            - args:
              - arg1
              - arg2
              - arg3
            - kwargs:
                kwarg_1: val_1
                kwarg_2: val_2
                kwarg_3: val_3
    """
    result = dict(comment=[], new_state=None, old_state=None, name=name, result=True)

    # Get defaults for each argument
    if path is None:
        path = name
    if args is None:
        args = []
    if kwargs is None:
        kwargs = {}

    func_ctx = await hub.idem.acct.ctx(
        path,
        profile=acct_profile,
        acct_data=acct_data or hub.idem.RUNS[ctx.run_name].get("acct_data"),
        hard_fail=True,
        validate=True,
    )

    # Other states have acct_data consumed already, but exec.run passes it forward to the exec module
    # ctx and ctx.acct have not been created yet from acct_data for the exec module.
    # At this point it is "acct_data" that needs to be removed rather than ctx or ctx.acct
    sanitized_kwargs = {k: v for k, v in kwargs.items() if k != "acct_data"}
    await hub.idem.event.put(
        profile="idem-state",
        body={name: {path: sanitized_kwargs}},
        tags={
            "ref": f"exec.{path}",
            "type": "state-pre",
            "acct_details": func_ctx.acct_details,
        },
    )

    # Report the cli command of the exec module being called
    cli_call = f"idem exec {path} --acct-profile={acct_profile}"
    if args:
        cli_call += " " + " ".join(args)
    if kwargs:
        cli_call += " " + " ".join(f"{k}={v}" for k, v in kwargs.items())
    if "name" not in kwargs:
        try:
            resolved = hub.idem.ex.resolve(path)
            params = resolved["params"]
            if "name" in params:
                kwargs["name"] = name
        except AttributeError:
            ...

    result["comment"] += [cli_call]

    # Get the acct_data from the current run
    acct_data = acct_data or hub.idem.RUNS[ctx.run_name]["acct_data"]

    # Run the exec module!
    try:
        ret = await hub.idem.ex.run(
            path=path,
            args=args,
            kwargs=kwargs,
            acct_data=acct_data,
            acct_profile=acct_profile,
            rerun_data=ctx.get("rerun_data"),
        )
        result["result"] &= ret.result
        result["new_state"] = ret.ret
        result["rerun_data"] = ret.get("rerun_data")
        # Avoid ESM by reporting changes directly
        result["changes"] = {"new": ret.ret}
        if ret.comment:
            if isinstance(ret.comment, List):
                result["comment"] += ret.comment
            elif isinstance(ret.comment, Tuple):
                result["comment"] += list(ret.comment)
            else:
                result["comment"].append(ret.comment)
    except Exception as e:
        # Float up the errors
        result["result"] = False
        result["comment"] += [f"{e.__class__.__name__}: {e}"]

    # Fire an event using the exec module's path instead of the exec.run
    await hub.idem.event.put(
        profile="idem-state",
        body=result,
        tags={
            "ref": f"exec.{path}",
            "type": "state-post",
            "acct_details": func_ctx.acct_details,
        },
    )
    return result


def is_pending(hub, ret):
    """
    Pending implementation of exec.run
    Pending if the 'result' is False.

    :return: True if reconciliation is required. If result is True do not reconcile.
    """
    return not ret["result"]



# NOTES
# This is a VERY simple outputter for idem, it does not do everything the
# Salt highstate outputter does, and nor should it! This outputter should
# not become hyper complicated, things like terse should be another
# outputter, this should really just get things like errors added
from typing import Dict

try:
    import colorama

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


def display(hub, data):
    """
    Display the data from an idem run
    """
    if not isinstance(data, dict):
        return hub.output.nested.display(data)

    endc = colorama.Fore.RESET
    strs = []
    fun_count = {}
    for tag in data:
        ret = data[tag]
        comps = tag.split("_|-")
        state = comps[0]
        id_ = comps[1]
        fun = comps[3]
        result = ret.get("result")
        comment = ret.get("comment", [])

        # when the current chunk's execution is halted due to recreation flow, we can skip that output from display.
        if comment and "will be recreated" in comment:
            hub.log.debug(
                f"The output of state {id_} will be skipped from display, as it just states that resource will be recreated."
            )
            continue

        changes = hub.output.nested.display(ret.get("changes", {}))
        if result is True and changes:
            tcolor = colorama.Fore.CYAN
        elif result is True:
            tcolor = colorama.Fore.GREEN
        elif result is None:
            tcolor = colorama.Fore.LIGHTYELLOW_EX
        elif result is False:
            tcolor = colorama.Fore.LIGHTRED_EX
        else:
            tcolor = colorama.Fore.RESET

        strs.append(f"{tcolor}--------{endc}")
        strs.append(f"{tcolor}      ID: {id_}{endc}")
        strs.append(f"{tcolor}Function: {state}.{fun}{endc}")
        strs.append(f"{tcolor}  Result: {result}{endc}")
        strs.append(f"{tcolor} Comment: {comment}{endc}")
        strs.append(f"{tcolor} Changes:\n{changes}{endc}")

        # Calculate counts for each function and result
        _increment_count(fun_count, fun, result)

        # Get granularity for successful result
        if result:
            if not ret.get("old_state"):
                _increment_count(fun_count, fun, "no_old_state")
            elif ret["changes"]:
                _increment_count(fun_count, fun, "with_changes")

    strs = strs + _format_fun_counts(fun_count)
    return "\n".join(strs)


def _format_fun_counts(fun_map: Dict[str, Dict[str, int]]) -> []:
    # Format counts for each function
    # Sample output:
    #  present: 1 successful
    #  present: 2 failed
    strs = ["\n"]

    for fun, result_and_count in fun_map.items():
        if result_and_count.get(True, 0) > 0:
            if fun == "present":
                if result_and_count.get("no_old_state", 0) > 0:
                    strs.append(
                        f"{fun}: {result_and_count['no_old_state']} created successfully"
                    )
                if result_and_count.get("with_changes", 0) > 0:
                    strs.append(
                        f"{fun}: {result_and_count['with_changes']} updated successfully"
                    )
                no_op_count = (
                    result_and_count[True]
                    - result_and_count.get("no_old_state", 0)
                    - result_and_count.get("with_changes", 0)
                )
                if no_op_count > 0:
                    strs.append(f"{fun}: {no_op_count} no-op")

            elif fun == "absent":
                if result_and_count.get("no_old_state", 0) > 0:
                    strs.append(f"{fun}: {result_and_count['no_old_state']} no-op")
                deletion_count = result_and_count[True] - result_and_count.get(
                    "no_old_state", 0
                )
                if deletion_count > 0:
                    strs.append(f"{fun}: {deletion_count} deleted successfully")

            else:
                strs.append(f"{fun}: {result_and_count[True]} successful")

        if result_and_count.get(False, 0) > 0:
            strs.append(f"{fun}: {result_and_count[False]} failed")

    return strs


def _increment_count(map, key1, key2):
    if map.get(key1) is None:
        map[key1] = {key2: 1}
    elif map[key1].get(key2) is None:
        map[key1][key2] = 1
    else:
        map[key1][key2] += 1



from typing import Any
from typing import Dict
from typing import List
from typing import Set


async def sig_apply(
    hub,
    name: str,
    state: Dict[str, Any],
    sls_ref: str,
    cfn: str,
    resolved: Set[str],
) -> List[str]:
    """
    :param hub:
    :param name: The state run name
    :param state: A rendered block from the sls
    :param sls_ref: A reference to another sls within the given sources
    :param cfn: The cache file name, or the location of sls within the given sources
    :param resolved: a set of refs that have already been resolved
    """



from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Set

KEYWORDS = ["order"]


async def apply(
    hub,
    name: str,
    state: Dict[str, Any],
    sls_ref: str,
    cfn: str,
    resolved: Set[str],
) -> List[str]:
    """
    Take a state and apply the iorder system

    :param hub:
    :param name: The state run name
    :param state: A rendered block from the sls
    :param sls_ref: A reference to another sls within the given sources
    :param cfn: The cache file name, or the location of sls within the given sources
    :param resolved: a set of refs that have already been resolved
    """
    for id_ in hub.idem.resolve.init.iter(state):
        if isinstance(state[id_], str):
            continue
        if not isinstance(state[id_], Iterable):
            continue
        for s_dec in state[id_]:
            if not isinstance(s_dec, str):
                # PyDSL OrderedDict?
                continue

            if not isinstance(state[id_], Dict):
                # Include's or excludes as lists?
                continue
            if not isinstance(state[id_][s_dec], List):
                # Bad syntax, let the verify seq pick it up later on
                continue

            found = False
            if s_dec.startswith("_"):
                continue

            for arg in state[id_][s_dec]:
                if isinstance(arg, Dict):
                    if len(arg) > 0:
                        if next(iter(arg)) == "order":
                            found = True
            if not found:
                if not isinstance(state[id_][s_dec], List):
                    # quite certainly a syntax error, managed elsewhere
                    continue

                state[id_][s_dec].append({"order": hub.idem.RUNS[name]["iorder"]})
                hub.idem.RUNS[name]["iorder"] += 1
    return []



import re

ARG_REFERENCE_REGEX_PATTERN = re.compile(r"\${[^\${}]+}")


def reconcile(hub, high):
    """
    Parse high data and convert argument binding references like ${cloud:state:property_path} to arg_bind requisites

    ex:
    state1:
     cloud1.state_def:
      - parameters:
        arg1: ${cloud2:state2:property_path}
      - arg_bind:
       - cloud2:
        - state2:
          - property_path: arg1
    """
    errors = []
    # TODO: Add debug logging
    # Iterate over yield arguments and check if the argument value contains references
    for id_, body, state, run, arg_key, arg_value in hub.idem.tools.iter_high_leaf_args(
        high
    ):
        if (
            arg_value
            and isinstance(arg_value, str)
            and ARG_REFERENCE_REGEX_PATTERN.search(arg_value)
        ):
            for arg_ref in ARG_REFERENCE_REGEX_PATTERN.findall(arg_value):
                ref_chain_list = str(arg_ref[2 : arg_ref.index("}")]).split(":")

                if len(ref_chain_list) < 3:
                    # TODO: Find out why this doesn't fail the execution
                    errors.append(
                        f'Argument reference {arg_ref} for state "{id_}" is not properly formatted. '
                        f"Argument reference format is ${{referenced_cloud:referenced_state:argument_path}}."
                    )
                    continue

                ref_cloud = ref_chain_list.pop(0)
                ref_state = ref_chain_list.pop(0)

                # This code is executed for every idem iteration.
                # Make sure, arg_bind requisites don't get duplicated.
                arg_bind_req = _get_or_add_arg(high.get(id_).get(state), "arg_bind", [])
                arg_bind_cloud = _get_or_add_arg(arg_bind_req, ref_cloud, [])
                arg_bind_state = _get_or_add_arg(arg_bind_cloud, ref_state, [])
                _get_or_add_arg(arg_bind_state, ":".join(ref_chain_list), arg_key)

    return high, errors


def _get_or_add_arg(arg_list, arg_name, arg_value):
    for arg in arg_list:
        # Expected format is a list of dictionaries
        if isinstance(arg, dict) and arg.get(arg_name, None):
            # The same referenced state property can be bound to multiple arguments of the current state.
            # If the argument list contains expected argument and argument value is string,
            # make sure it is set to the argument value, otherwise add new argument for the value.
            if isinstance(arg_value, str) and isinstance(arg.get(arg_name), str):
                if arg.get(arg_name) == arg_value:
                    return arg.get(arg_name)
            else:
                return arg.get(arg_name)

    arg_state = {arg_name: arg_value}
    arg_list.append(arg_state)
    return arg_state.get(arg_name)



from typing import Any
from typing import Dict
from typing import List
from typing import Set

KEYWORDS = ["META"]


async def apply(
    hub,
    name: str,
    state: Dict[str, Any],
    sls_ref: str,
    cfn: str,
    resolved: Set[str],
) -> List[str]:
    """
    :param hub:
    :param name: The state run name
    :param state: A rendered block from the sls
    :param sls_ref: A reference to another sls within the given sources
    :param cfn: The cache file name, or the location of sls within the given sources
    :param resolved: a set of refs that have already been resolved
    """
    if "META" in state:
        raw_meta = state.pop("META")
        hub.idem.RUNS[name]["meta"]["SLS"][sls_ref] = raw_meta
    for id_ in state:
        if isinstance(state[id_], Dict):
            if "META" in state[id_]:
                ref = f"{sls_ref}.{id_}"
                raw_meta = state[id_].pop("META")
                hub.idem.RUNS[name]["meta"]["ID_DECS"][ref] = raw_meta
    return []



from typing import Any
from typing import Dict


def check(
    hub,
    name: str,
    ctx: Dict[str, Any],
    condition: Any,
    reqret: Dict[str, Any],
    chunk: Dict[str, Any],
    managed_state: Dict[str, Any],
    execution_seq: Dict,
) -> Dict[str, Any]:
    """
    Make sensitive data from sls blocks to be available in Idem RUNS global variable.
    """
    if condition != "sensitive":
        return {"errors": [f'"{condition}" is not sensitive requisite.']}
    if "sensitive" in chunk:
        if not isinstance(chunk["sensitive"], list):
            return {
                "errors": [
                    f'sensitive requisite should contain a list of sensitive parameters, not {chunk["sensitive"]}'
                ]
            }
        tag = hub.idem.tools.gen_chunk_func_tag(chunk)
        if hub.idem.RUNS[name].get("sensitive"):
            hub.idem.RUNS[name]["sensitive"].update({tag: chunk["sensitive"]})
        else:
            hub.idem.RUNS[name]["sensitive"] = {tag: chunk["sensitive"]}
    return {"errors": []}



import io
import os
import tarfile
from pathlib import Path
from typing import ByteString
from typing import Tuple

import msgpack
from cryptography.fernet import Fernet

MIMETYPES = [".epkg", "epkg"]

"""
Process encrypted archive source
"""

__virtualname__ = "epkg"


def __init__(hub):
    hub.source.file.ACCT = ["file"]


async def cache(
    hub, ctx, protocol: str, source: str, location: str
) -> Tuple[str, ByteString]:
    # name of epkg and key name (to decrypt the package) should match
    source_path = Path(source)
    key = source_path.stem
    # Normalize the location so that it can be matched with member name
    location = location.replace(os.sep, ".")
    f = Fernet(ctx.acct.get(key))

    with open(source, "rb") as rfh:
        data = msgpack.loads(rfh.read())

    decrypted_data = f.decrypt(data["token"])
    raw = io.BytesIO(decrypted_data)

    with tarfile.open(fileobj=raw, mode="r:xz") as tar:
        members = tar.getmembers()

        for member in members:
            # The inner member.name will always have slashes regardless of OS
            member_name = member.name.replace("/", ".")
            if member_name.endswith(location):
                # Use the unmodified member name
                location = member.name
                break
        else:
            # Never reached the break
            return "", b""
        file = tar.extractfile(location)

        if file:
            # Use <source>/<sls file> as the unique sls_ref
            return f"{source}/{location}", file.read()
    return "", b""



import json
from pathlib import Path
from typing import Dict

from jinja2 import Template


async def write(
    hub,
    ctx,
    parameters: dict,
    file_name: str = None,
    template: str = None,
    template_file_name: str = None,
) -> Dict[str, Dict]:
    """
    This is a resource that produces output that is recorded in a file.
    If the output file exists it will be overwritten.
    If a jinja template is provided it should accept the "parameters" dictionary and
    the rendered output will be saved to the file.
    Without a Jinja template the output file includes the "parameters" in a json format.
    Jinja template can be passed raw or within a template file

    Sample template:
    {% for resource_name, resource_id in parameters.items() %}
        Created resource: {{ resource_name }} {{ resource_id }}
    {% endfor %}

    Args:
        file_name(Text): The name of the output file with a full path
        parameters(Dict): dictionary of parameters
        template(Text, Optional): Embedded Jinja template within {% raw  %} {% endraw  %}
        template_file_name(Text, Optional): Full path to a jinja template file
            either template or template_file should be specified

    Returns:
        {"result": True|False, "comment": "A message", "old_state": { file_name: content },
        "new_state": { file_name: content }}

    Request Syntax:
        [output-id]:
          data.write:
          - file_name: 'string'
          - template: 'string'
          - parameters: Dict

    Examples:
        .. code-block:: sls

            my-output-file:
                data.write:
                - file_name: "/home/files/OUTPUT.json"
                - template_file_name: "/home/files/template.txt"
                - parameters:
                    resource-test-1.id: ${resource-type:resource-test-1:id}
                    resource-test-1.name: ${resource-type:resource-test-1:name}

            my-output-file:
                data.write:
                - file_name: "/home/files/OUTPUT.json"
                - template: '{% raw  %}
                                {% for resource_name, resource_id in parameters.items() %}
                                    Created resource {{ resource_name }}  {{ resource_id }}
                                {% endfor %}
                            {%  endraw %}'
                - parameters:
                    resource-test-1.id: ${resource-type:resource-test-1:id}
                    resource-test-1.name: ${resource-type:resource-test-1:name}

    """
    result = dict(comment=(), old_state={}, new_state={}, name=file_name, result=True)

    if file_name is None or str(file_name) == 0:
        result["result"] = False
        result["comment"] = ("File name is required (`file_name`).",)
        return result

    if ctx.get("test", False):
        result["comment"] = (f"Would write to file '{file_name}'.",)
        return result

    try:
        # If the output file exists, initialize 'old_state'
        outfile = Path(file_name)
        if outfile.is_file():
            hub.log.debug(
                f"File {file_name} already exists. Its content will be overwritten."
            )
            outfile = open(file_name)
            content = outfile.read()
            if content is not None and len(content) > 0:
                result["old_state"] = {file_name: content}

        output_content = ""
        # Jinja template that accepts a dictionary "parameters"
        if template_file_name is not None:
            template_file = Path(template_file_name)
            if template_file.is_file() is False:
                msg = f"Template file not found: {template_file_name}"
                hub.log.debug(msg)
                result["result"] = False
                result["comment"] = (msg,)
                return result
            else:
                with open(template_file) as curr_template_file:
                    tm = Template(curr_template_file.read())
                    output_content = tm.render(parameters=parameters)
        # Embedded Jinja template
        elif template is not None:
            tm = Template(template)
            output_content = tm.render(parameters=parameters)
        else:
            # No template
            output_content = json.dumps(parameters, indent=4)

        # Writing to the file
        with open(file_name, "w") as outfile:
            outfile.write(output_content)
            outfile.close()

    except FileNotFoundError:
        result["result"] = False
        result["comment"] = (
            f"File(s) not found. File: '{file_name}', template: '{template_file_name}'",
        )
        return result

    result["comment"] = f"Wrote to file '{file_name}'."
    result["new_state"] = {file_name: output_content}
    return result



from typing import Any
from typing import Dict


async def run(
    hub,
    plugin: str = "basic",
    pending_plugin: str = "default",
    name: str = None,
    apply_kwargs: Dict[str, Any] = None,
):
    if apply_kwargs is None:
        apply_kwargs = {}
    apply_kwargs["name"] = name
    try:
        ret = await hub.reconcile[plugin].loop(
            pending_plugin,
            name=name,
            apply_kwargs=apply_kwargs,
        )
    finally:
        hub.idem.state.update_status(name, hub.idem.state.Status.FINISHED)

    return ret



"""
If an exec module plugin implements this contract,
the same ref can be used to call the state functions "present", "absent", and "describe".
"""
from typing import Any
from typing import Dict


def sig_get(hub, ctx, name: str, *args, **kwargs) -> Dict[str, Any]:
    return {"result": True | False, "comment": "", "ret": None}


def sig_list(hub, ctx, *args, **kwargs) -> Dict[str, Any]:
    return {"result": True | False, "comment": "", "ret": None}


def sig_create(hub, ctx, name: str, *args, **kwargs) -> Dict[str, Any]:
    return {"result": True | False, "comment": "", "ret": None}


def sig_update(hub, ctx, name: str, *args, **kwargs) -> Dict[str, Any]:
    return {"result": True | False, "comment": "", "ret": None}


def sig_delete(hub, ctx, name: str, *args, **kwargs) -> Dict[str, Any]:
    return {"result": True | False, "comment": "", "ret": None}



import contextlib
import pathlib
import traceback
import uuid
from collections.abc import MutableMapping
from typing import Any
from typing import Callable
from typing import Dict


def __init__(hub):
    # This enables integration tests to prevent deletion of the cache_file
    # When Idem is called from the CLI, this field is always overridden by idem/idem/init.py
    hub.idem.managed.KEEP_CACHE_FILE = True


@contextlib.asynccontextmanager
async def context(
    hub,
    run_name: str,
    cache_dir: str,
    esm_plugin: str = "local",
    esm_profile: str = "default",
    acct_file: str = None,
    acct_key: str = None,
    acct_blob: str = None,
    acct_data: Dict[str, Any] = None,
    serial_plugin: str = "msgpack",
):
    """
    Only allow one instance of this run within the context of the enforced state manager
    """
    cache_dir = pathlib.Path(cache_dir)
    esm_cache_file = (
        cache_dir / "esm" / "cache" / f"{run_name}-{uuid.uuid4()}.{serial_plugin}"
    )
    esm_cache_file.parent.mkdir(parents=True, exist_ok=True)
    ctx = await hub.idem.acct.ctx(
        f"esm.{esm_plugin}",
        profile=esm_profile,
        acct_key=acct_key,
        acct_file=acct_file,
        acct_blob=acct_blob,
        acct_data=acct_data,
    )
    # If no profile was specified then use the default profile
    if esm_plugin == "local" and not ctx.acct:
        hub.log.debug("Using the default local ESM profile")
        ctx = await hub.idem.acct.ctx(
            "esm.local",
            profile=None,
            acct_data={
                "profiles": {
                    "esm.local": {
                        None: {
                            "run_name": run_name,
                            "cache_dir": cache_dir,
                            "serial_plugin": serial_plugin,
                        }
                    }
                }
            },
        )

    exception = None
    # Enter the context of the Enforced State Manager
    # Do this outside of the try/except so that exceptions don't cause unintentional release of lock in exit
    try:
        handle = await hub.esm[esm_plugin].enter(ctx)
    except Exception as e:
        raise RuntimeError(
            f"Fail to enter enforced state management: {e.__class__.__name__}: {e}"
        )
    try:
        # Get the current state from the context
        state: Dict[str, Any] = await hub.esm[esm_plugin].get_state(ctx) or {}
        cache_state = hub.idem.managed.file_dict(
            cache_file=str(esm_cache_file), data=state, serial_plugin=serial_plugin
        )
        # The cache_state can be interacted with like a regular dictionary, but the file is always up-to-date
        yield cache_state
        # update the enforced state from the cache
        data = cache_state.data
        await hub.esm[esm_plugin].set_state(ctx, data)
        # Remove the cache file, everything has been stored in the final destination
        if not hub.idem.managed.KEEP_CACHE_FILE:
            hub.log.debug("Removing the temporary local ESM cache")
            esm_cache_file.unlink()
    except Exception as e:
        exception = e
        raise
    finally:
        # Exit the context of the Enforced State Manager
        try:
            if exception is not None:
                # This exception can be raised by anything while Idem state is running, so for best debugging practice,
                # we log the stacktrace in debug mode
                hub.log.debug(traceback.format_exc())
            await hub.esm[esm_plugin].exit_(ctx, handle, exception)
        except Exception as e:
            raise RuntimeError(
                f"Fail to exit enforced state management: {e.__class__.__name__}: {e}"
            )


def file_dict(hub, cache_file: str, data: Dict = None, serial_plugin: str = "msgpack"):
    return FileDict(
        cache_file=cache_file,
        serializer=hub.serial[serial_plugin].dump,
        deserializer=hub.serial[serial_plugin].load,
        data=data,
    )


class FileDict(MutableMapping):
    """
    Any time there is a change to this dictionary, it will immediately be reflected in a cache file
    """

    def __init__(
        self,
        cache_file: str,
        deserializer: Callable,
        serializer: Callable,
        data: Dict = None,
    ):
        self.deserialize = deserializer
        self.serialize = serializer
        if data is None:
            data = {}
        self.file = pathlib.Path(cache_file)
        with self.file.open("wb+") as fh:
            byte_data: bytes = self.serialize(data)
            fh.write(byte_data)

    @property
    def data(self):
        with self.file.open("rb+") as fh:
            return self.deserialize(fh.read())

    def __iter__(self):
        return iter(self.data)

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, k):
        return self.data[k]

    def __delitem__(self, v):
        data = self.data
        data.pop(v)
        with self.file.open("wb+") as fh:
            byte_data: bytes = self.serialize(data)
            fh.write(byte_data)

    def __setitem__(self, k, v):
        data = self.data
        data[k] = v
        with self.file.open("wb+") as fh:
            byte_data: bytes = self.serialize(data)
            fh.write(byte_data)



import datetime
from typing import Any
from typing import Dict

import pop.loader


# These are keywords passed to state module functions which are to be used
# by idem in this state module and not on the actual state module function
STATE_REQUISITE_KEYWORDS = frozenset(
    [
        "onchanges",
        "onchanges_any",
        "onfail",
        "onfail_any",
        "onfail_all",
        "onfail_stop",
        "prereq",
        "prerequired",
        "watch",
        "watch_any",
        "require",
        "require_any",
        "listen",
        "arg_bind",
        "sensitive",
        "ignore_changes",
        "recreate_on_update",
    ]
)
STATE_REQUISITE_IN_KEYWORDS = frozenset(
    [
        "onchanges_in",
        "onfail_in",
        "prereq_in",
        "watch_in",
        "require_in",
        "listen_in",
    ]
)
STATE_RUNTIME_KEYWORDS = frozenset(
    [
        "fun",
        "state",
        "check_cmd",
        "failhard",
        "onlyif",
        "unless",
        "retry",
        "order",
        "parallel",
        "prereq",
        "prereq_in",
        "prerequired",
        "reload_modules",
        "reload_grains",
        "reload_pillar",
        "runas",
        "runas_password",
        "fire_event",
        "saltenv",
        "use",
        "use_in",
        "__run_name",
        "__env__",
        "__sls__",
        "__id__",
        "__orchestration_jid__",
        "__pub_user",
        "__pub_arg",
        "__pub_jid",
        "__pub_fun",
        "__pub_tgt",
        "__pub_ret",
        "__pub_pid",
        "__pub_tgt_type",
        "__prereq__",
    ]
)

STATE_INTERNAL_KEYWORDS = STATE_REQUISITE_KEYWORDS.union(
    STATE_REQUISITE_IN_KEYWORDS
).union(STATE_RUNTIME_KEYWORDS)


def get_func(hub, name, chunk, fun=None):
    """
    Given the runtime name and the chunk in question, determine what function
    on the hub that can be run
    """
    if fun is None:
        fun = chunk["fun"]
    s_ref = chunk["state"]

    # Check if an auto_state exists for this ref
    try:
        if "auto_state" in hub.exec[s_ref].__contracts__:
            chunk["exec_mod_ref"] = s_ref
            return hub.states.auto_state[fun]
    except AttributeError or TypeError:
        ...
    for sub in hub.idem.RUNS[name]["subs"]:
        test = f"{sub}.{s_ref}.{fun}"
        try:
            func = getattr(hub, test)
        except AttributeError:
            continue
        if isinstance(func, pop.loader.LoadedMod):
            continue
        if func is None:
            continue
        return func
    return None


async def run(
    hub, name, ctx, low, seq_comp, running, run_num, managed_state, execution_seq
):
    """
    All requisites have been met for this low chunk.
    """
    chunk = seq_comp["chunk"]
    tag = hub.idem.tools.gen_chunk_func_tag(chunk)
    esm_tag = hub.idem.tools.gen_chunk_esm_tag(chunk)
    start_time = datetime.datetime.now()
    skip_check = ["resolver"]
    rdats = {}
    running[tag] = {
        "tag": tag,
        "name": chunk["name"],
        "__id__": chunk["__id__"],
        "changes": {},
        "new_state": None,
        "old_state": None,
        "comment": None,
        "rerun_data": None,
        "result": False,
        "esm_tag": esm_tag,
        "__run_num": run_num,
        "start_time": str(start_time),
        "total_seconds": 0,
        "sls_meta": hub.idem.RUNS[name]["meta"],
    }

    # setting the "sls_run_id" attribute in running["tag"]
    # so that we can identify later that this chunk is part of a sls.run group
    if "sls_run_id" in chunk:
        running[tag]["sls_run_id"] = chunk["sls_run_id"]

    ctx["tag"] = tag
    ctx["rerun_data"] = chunk.get("rerun_data")

    errors = seq_comp.get("errors", [])
    for reqret in seq_comp.get("reqrets", []):
        req = reqret["req"]
        if req not in rdats:
            rdats[req] = []
        rules = hub.idem.RMAP[req]
        for rule in rules:
            if rule in skip_check:
                continue
            if hasattr(hub.idem.rules, rule):
                rdat = hub.idem.rules[rule].check(
                    name,
                    ctx,
                    rules[rule],
                    reqret,
                    chunk,
                    managed_state,
                    execution_seq,
                )
                rdats[req].append(rdat)

    # Get the function to be called as early as possible
    func = hub.idem.rules.init.get_func(name, chunk)
    ref = None
    if func:
        ref = f"{func.ref}.{func.__name__}"

    errors += hub.idem.resolver.init.resolve(rdats)
    if errors:
        running[tag]["comment"] = "\n".join(errors)
        running[tag]["total_seconds"] = (
            datetime.datetime.now() - start_time
        ).total_seconds()
        await hub.idem.event.put(
            profile="idem-run",
            body=running[tag],
            tags={
                "ref": ref,
                "type": "state-result",
                "acct_details": None,
            },
        )
        return

    # if the requisite `recreate_on_update` is processed and `create_before_destroy=false`
    # we will halt the current chunk from execution as the chunks
    # for destroying old resource and creating new resource will be executed in next run.

    if chunk.get("halt_current_execution", False):
        msg = f"The resource {chunk['__id__']} will be recreated."
        hub.log.debug(msg)
        running[tag]["comment"] = msg
        running[tag]["result"] = True
        running[tag]["total_seconds"] = (
            datetime.datetime.now() - start_time
        ).total_seconds()
        await hub.idem.event.put(
            profile="idem-run",
            body=running[tag],
            tags={
                "ref": ref,
                "type": "state-result",
                "acct_details": None,
            },
        )

        # Update the result for reconciliation
        hub.idem.RUNS[name]["running"][tag]["comment"] = msg
        hub.idem.RUNS[name]["running"][tag]["result"] = True
        return

    await hub.idem.event.put(
        profile="idem-chunk",
        body=chunk,
        tags={
            "ref": ref,
            "type": "state-chunk",
        },
    )
    if func is None:
        running[tag]["comment"] = (
            f'Could not find function to enforce {chunk["state"]}. '
            f"Please make sure that the corresponding plugin is loaded."
        )
        running[tag]["total_seconds"] = (
            datetime.datetime.now() - start_time
        ).total_seconds()
        await hub.idem.event.put(
            profile="idem-run",
            body=running[tag],
            tags={
                "ref": ref,
                "type": "state-result",
                "acct_details": None,
            },
        )
        return
    chunk["ctx"] = ctx
    chunk = await hub.idem.mod.init.modify(name, chunk)
    call_kwargs = hub.idem.tools.format_call(
        fun=func,
        data=chunk,
        ignore_changes=chunk.get("ignore_changes"),
        expected_extra_kws=STATE_INTERNAL_KEYWORDS,
        enforced_state=hub.idem.tools.get_enforced_state(chunk, managed_state),
    )
    for req, rlist in rdats.items():
        for rdat in rlist:
            if "pre" in rdat:
                ret = rdat["pre"](**call_kwargs)
                await hub.pop.loop.unwrap(ret)
    # This is when the state is actually called
    ret = func(**call_kwargs)
    ret = await hub.pop.loop.unwrap(ret)

    # Update ESM under the following conditions:
    # ESM will be updated when Idem ran with "refresh" subcommand (which implies the `test` flag).
    # Or, when Idem state is not running with --test
    # When the above rule is met, one of the following conditions need to be met:
    # - "force_save" is True (This implies the plugin state has insisted Idem to save a resource state)
    # - "result" is True (This implies the Idem state operation is successful)
    # - "old_state" doesn't exist but "new_state" exists. This implies a creation operation during the
    #    Idem state run. The "new_state" will be saved even if the operation failed ("result" is False).
    if (
        (getattr(hub, "SUBPARSER", None) == "refresh" or not ctx.get("test"))
        and (
            ret.get("result")
            or ret.get("force_save", False)
            or ((not ret.get("old_state")) and ret.get("new_state"))
        )
        and not chunk["ctx"].get("skip_esm", False)
    ):
        new_state = ret.get("new_state")
        if new_state:
            # Changes were made, update the cache
            managed_state[esm_tag] = new_state
        else:
            # The successful new_state shows a deleted resource, remove it from the cache
            managed_state.pop(esm_tag, None)

    # Remove force_save from ret as it has been used and does not need to be kept in ret
    ret.pop("force_save", None)

    for req, rlist in rdats.items():
        for rdat in rlist:
            if "post" in rdat:
                ret = rdat["post"](**call_kwargs)
                ret = await hub.pop.loop.unwrap(ret)

    # Update the result for reconciliation
    # if the current executing chunk is for recreating a resource, then update it in ret also
    if chunk.get("recreation_flow", False):
        ret["recreation_flow"] = True

    hub.idem.RUNS[name]["running"][tag].update(ret)

    running[tag].update(ret)
    running[tag]["total_seconds"] = (
        datetime.datetime.now() - start_time
    ).total_seconds()

    await hub.idem.event.put(
        profile="idem-run",
        body=running[tag],
        tags={
            "ref": ref,
            "type": "state-result",
            "acct_details": chunk["ctx"].get("acct_details"),
        },
    )


def check(
    hub,
    name: str,
    ctx: Dict[str, Any],
    condition: Any,
    reqret: Dict[str, Any],
    chunk: Dict[str, Any],
    managed_state: Dict[str, Any],
    execution_seq: Dict,
) -> Dict[str, Any]:
    ...



"""
Plugin to skip state management
"""
from typing import Any
from typing import Dict


async def get_state(hub, ctx) -> Dict[str, Any]:
    return {}


async def set_state(hub, ctx, state: Dict[str, Any]):
    ...


async def enter(hub, ctx):
    ...


async def exit_(hub, ctx, handle, exception: Exception):
    ...



def _validate_extracted_params_recursively(state, extracted_params, meta, dot):
    if isinstance(extracted_params, dict):
        for key, value in extracted_params.items():
            if meta is None or key not in meta:
                yield (state, f"{dot}{key}")
                next_meta = None
            elif meta[key] is not None and "properties" in meta[key]:
                next_meta = meta[key]["properties"]
            else:
                next_meta = None

            for state, param in _validate_extracted_params_recursively(
                state, value, next_meta, "."
            ):
                yield (state, f"{dot}{key}{param}")
    elif isinstance(extracted_params, list):
        for item in extracted_params:
            if item is None or isinstance(item, str):
                continue

            if meta is None or key not in meta:
                next_meta = None
            elif meta[key] is not None and "items" in meta[key]:
                next_meta = meta[key]["items"]
            else:
                next_meta = None

            for state, param in _validate_extracted_params_recursively(
                state, item, next_meta, "."
            ):
                yield (state, f"[]{param}")


def _validate_extracted_params(parameters, meta):
    for state, extracted_params in parameters.items():
        if meta is None or state not in meta or "params" not in meta[state]:
            state_meta = None
        else:
            state_meta = meta[state]["params"]

        for state, param in _validate_extracted_params_recursively(
            state, extracted_params, state_meta, ""
        ):
            yield (state, param)


def stage(hub, name):
    hub.idem.RUNS[name]["warnings"] = {"GLOBAL": {}, "ID_DECS": {}}

    if (
        "parameters" not in hub.idem.RUNS[name]
        or "ID_DECS" not in hub.idem.RUNS[name]["parameters"]
    ):
        return

    if (
        "meta" not in hub.idem.RUNS[name]
        or "ID_DECS" not in hub.idem.RUNS[name]["meta"]
    ):
        all_meta = None
    else:
        all_meta = hub.idem.RUNS[name]["meta"]["ID_DECS"]

    all_warnings = {}

    for state, missing_param in _validate_extracted_params(
        hub.idem.RUNS[name]["parameters"]["ID_DECS"], all_meta
    ):
        if state not in all_warnings:
            all_warnings[state] = {"params_meta_missing": []}
        all_warnings[state]["params_meta_missing"].append(missing_param)

    for state, warnings in all_warnings.items():
        hub.idem.RUNS[name]["warnings"]["ID_DECS"][state] = warnings



from typing import Dict
from typing import List


def sig_get_refs(hub, *, sources: List[str], refs: List[str]) -> Dict[str, List[str]]:
    """
    :param hub:
    :param sources: sls-sources or params-sources
    :param refs: References to sls within the given sources
    """


async def sig_gather(hub, name: str, *sls):
    """
    :param hub:
    :param name: The state run name
    :param sls: sls locations within sources
    """



from typing import Any
from typing import Dict


def define(hub) -> Dict[str, Any]:
    """
    Return the definition used by the runtime to insert the conditions of the
    given requisite
    """
    return {
        "recreate_on_update": "recreate_on_update",
    }



def reconcile(hub, high, disabled_reqs=None):
    """
    Extend the data reference with requisite_in arguments
    """
    req_in = {
        "require_in",
        "watch_in",
        "onfail_in",
        "onchanges_in",
        "use",
        "use_in",
    }
    req_in_all = req_in.union(
        {"require", "watch", "onfail", "onfail_stop", "onchanges"}
    )
    extend = {}
    errors = []
    if disabled_reqs is None:
        disabled_reqs = []
    if not isinstance(disabled_reqs, list):
        disabled_reqs = [disabled_reqs]
    # Highdata iterator
    for id_, body, state, run, arg in hub.idem.tools.iter_high(high):
        # Iterator yields args
        if isinstance(arg, dict):
            # It is not a function, verify that the arg is a
            # requisite in statement
            if len(arg) < 1:
                # Empty arg dict
                # How did we get this far?
                continue
            # Split out the components
            key = next(iter(arg))
            if key not in req_in:
                continue
            if key in disabled_reqs:
                continue
            rkey = key.split("_")[0]
            items = arg[key]
            if isinstance(items, dict):
                # Formatted as a single req_in
                for _state, name in items.items():
                    # Not a use requisite_in
                    found = False
                    if name not in extend:
                        extend[name] = {}
                    if "." in _state:
                        errors.append(
                            f"Invalid requisite in {rkey}: {_state} for "
                            f'{name}, in SLS "{body["__sls__"]}". Requisites must '
                            f'not contain dots, did you mean "{_state[:_state.find(".")]}"?'
                        )
                        _state = _state.split(".")[0]
                    if _state not in extend[name]:
                        extend[name][_state] = []
                    extend[name]["__env__"] = body["__env__"]
                    extend[name]["__sls__"] = body["__sls__"]
                    for ind in range(len(extend[name][_state])):
                        if next(iter(extend[name][_state][ind])) == rkey:
                            # Extending again
                            extend[name][_state][ind][rkey].append({state: id_})
                            found = True
                    if found:
                        continue
                    # The rkey is not present yet, create it
                    extend[name][_state].append({rkey: [{state: id_}]})
            if isinstance(items, list):
                # Formed as a list of requisite additions
                hinges = []
                for ind in items:
                    if not isinstance(ind, dict):
                        # Malformed req_in
                        if ind in high:
                            _ind_high = [x for x in high[ind] if not x.startswith("__")]
                            ind = {_ind_high[0]: ind}
                        else:
                            found = False
                            for _id in iter(high):
                                for state in [
                                    state
                                    for state in iter(high[_id])
                                    if not state.startswith("__")
                                ]:
                                    for j in iter(high[_id][state]):
                                        if isinstance(j, dict) and "name" in j:
                                            if j["name"] == ind:
                                                ind = {state: _id}
                                                found = True
                            if not found:
                                continue
                    if len(ind) < 1:
                        continue
                    pstate = next(iter(ind))
                    pname = ind[pstate]
                    if pstate == "sls":
                        # Expand hinges here
                        hinges = find_sls_ids(pname, high)
                    else:
                        hinges.append((pname, pstate))
                    for tup in hinges:
                        name, _state = tup
                        if key == "use_in":
                            # Add the running states args to the
                            # use_in states
                            ext_ids = find_name(name, _state, high)
                            for ext_id, _req_state in ext_ids:
                                if not ext_id:
                                    continue
                                ext_args = state_args(ext_id, _state, high)
                                if ext_id not in extend:
                                    extend[ext_id] = {}
                                if _req_state not in extend[ext_id]:
                                    extend[ext_id][_req_state] = []
                                ignore_args = req_in_all.union(ext_args)
                                for arg in high[id_][state]:
                                    if not isinstance(arg, dict):
                                        continue
                                    if len(arg) != 1:
                                        continue
                                    if next(iter(arg)) in ignore_args:
                                        continue
                                    # Don't use name or names
                                    if next(arg.keys()) == "name":
                                        continue
                                    if next(arg.keys()) == "names":
                                        continue
                                    extend[ext_id][_req_state].append(arg)
                            continue
                        if key == "use":
                            # Add the use state's args to the
                            # running state
                            ext_ids = find_name(name, _state, high)
                            for ext_id, _req_state in ext_ids:
                                if not ext_id:
                                    continue
                                loc_args = state_args(id_, state, high)
                                if id_ not in extend:
                                    extend[id_] = {}
                                if state not in extend[id_]:
                                    extend[id_][state] = []
                                ignore_args = req_in_all.union(loc_args)
                                for arg in high[ext_id][_req_state]:
                                    if not isinstance(arg, dict):
                                        continue
                                    if len(arg) != 1:
                                        continue
                                    if next(iter(arg)) in ignore_args:
                                        continue
                                    # Don't use name or names
                                    if next(arg.keys()) == "name":
                                        continue
                                    if next(arg.keys()) == "names":
                                        continue
                                    extend[id_][state].append(arg)
                            continue
                        found = False
                        if name not in extend:
                            extend[name] = {}
                        if _state not in extend[name]:
                            extend[name][_state] = []
                        extend[name]["__env__"] = body.get("__env__", "base")
                        extend[name]["__sls__"] = body.get("__sls__", "base")
                        for ind in range(len(extend[name][_state])):
                            if next(iter(extend[name][_state][ind])) == rkey:
                                # Extending again
                                extend[name][_state][ind][rkey].append({state: id_})
                                found = True
                        if found:
                            continue
                        # The rkey is not present yet, create it
                        extend[name][_state].append({rkey: [{state: id_}]})
    high["__extend__"] = []
    for key, val in extend.items():
        high["__extend__"].append({key: val})
    req_in_high, req_in_errors = hub.idem.ccomps.extend.reconcile(high)
    errors.extend(req_in_errors)
    return req_in_high, errors



import copy
from typing import Iterable


def create(hub, iterable: Iterable, **kwargs) -> Iterable:
    """
    Create a progress bar that updates as the iterable is iterated.
    """
    # Only show this progress bar if the "--progress" flag was set
    enabled = hub.OPT.get("idem", {}).get("progress", False)
    if not enabled:
        return iterable

    progress_plugin = hub.OPT.get("idem", {}).get("progress_plugin", "tqdm")

    if not hasattr(hub.tool.progress, progress_plugin):
        hub.log.trace(f"No progress bar plugin '{progress_plugin}' is loaded")
        return iterable

    # Add other options to the progress bar from config
    progress_opts = copy.copy(hub.OPT.get("idem", {}).get("progress_options", {}))
    if not isinstance(progress_opts, dict):
        hub.log.debug(
            f"'hub.OPT.idem.progress_options' is not a dictionary: {type(progress_opts)}"
        )
        progress_opts = {}

    # Merge config options with explicitly passed options
    progress_opts.update(kwargs)

    progress_bar = hub.tool.progress[progress_plugin].create(
        iterable,
        **progress_opts,
    )

    return progress_bar


def update(hub, progress_bar: Iterable, **kwargs):
    """
    Update the progress bar through the progress plugin
    """
    # Only show this progress bar if the "--progress" flag was set
    enabled = hub.OPT.get("idem", {}).get("progress", False)
    if not enabled:
        return

    progress_plugin = hub.OPT.get("idem", {}).get("progress_plugin", "tqdm")

    if not hasattr(hub.tool.progress, progress_plugin):
        hub.log.trace(f"No progress bar plugin '{progress_plugin}' is loaded")
        return

    hub.tool.progress[progress_plugin].update(
        progress_bar,
        **kwargs,
    )



from typing import Any

import dict_tools.data as dict_data


class ExecReturn(dict_data.NamespaceDict):
    def __init__(
        self, result: bool, ret: Any, comment: Any = None, ref: str = "", **kwargs
    ):
        """
        Exec Returns must have the keys of "result", "ret", and "comment".
        Any other values can be added to the namespace
        """
        if not isinstance(result, bool):
            raise ValueError("Got a non boolean value for exec return `result`")
        if not isinstance(ref, str):
            raise ValueError("Got a non string value for exec return `ref`")

        super().__init__(result=result, ret=ret, comment=comment, ref=ref, **kwargs)

    def __bool__(self):
        return self.result



from typing import Any
from typing import Dict


def define(hub) -> Dict[str, Any]:
    """
    Define how the onchanges requisite should run
    """
    return {
        "result": True,
        "changes": True,
        "resolver": "any",
    }



import pathlib
import re
from typing import Any
from typing import Dict
from typing import List


def get_refs(hub, *, sources: List[str], refs: List[str]) -> Dict[str, List[str]]:
    """
    Determine where the sls sources are

    :param hub:
    :param sources: sls-sources or params-sources
    :param refs: References to sls within the given sources
    """
    sls_sources = []
    slses = []
    for ref in refs:
        hub.idem.sls_source.init.process(sls_sources, slses, ref)

    sls_sources.extend(sources)

    return {"sls_sources": sls_sources, "sls": slses}


def process(hub, sources: List[str], slses: List[str], ref: str):
    """
    If an SLS ref represents a local file, then make sure its location is added to sources

    :param sources: sls-sources or params-sources
    :param slses: sls locations within sources
    :param ref: A reference to an sls within the given sources
    """
    # Normalize the sls ref to be all dots
    normalized_ref = hub.idem.sls_source.init.normalize(ref)
    # Check if this is a local file path that exists
    de_normalized_ref = hub.idem.sls_source.init.de_normalize(ref)

    # Check if the sls input was a file
    path = pathlib.Path(de_normalized_ref)

    # If multiple SLS were passed with "init" as the final name, imply a source without the init to avoid collisions
    if path.stem == "init":
        path = path.parent

    for loc, _ in hub.idem.get.locations(ref):
        loc_path = pathlib.Path(loc)
        if loc_path.is_file():
            slses.append(normalized_ref)
            implied = f"file://{loc_path.parent.resolve().absolute()}"
            if implied not in sources:
                sources.append(implied)
                return
    else:
        # Input is not a file
        slses.append(normalized_ref)


async def gather(hub, name: str, *sls):
    """
    Gather the named sls references into the RUNS structure

    :param hub:
    :param name: The state run name
    :param sls: sls locations within sources
    """
    sources = hub.idem.RUNS[name]["sls_sources"]
    gather_data = await hub.idem.resolve.init.gather(name, *sls, sources=sources)
    await hub.idem.sls_source.init.update(name, gather_data)


async def update(hub, name: str, gather_data: Dict[str, Any]):
    """
    Update the RUNS dictionary with the output of the render process

    :param hub:
    :param name: The state run name
    :param gather_data: The output of hub.idem.resolve.init.gather/render

    """
    if gather_data.get("errors"):
        hub.idem.RUNS[name]["errors"] = gather_data["errors"]
        return
    if gather_data.get("blocks"):
        hub.idem.RUNS[name]["blocks"] = gather_data["blocks"]
    if gather_data.get("sls_refs"):
        hub.idem.RUNS[name]["sls_refs"] = gather_data["sls_refs"]
    if gather_data.get("resolved"):
        hub.idem.RUNS[name]["resolved"].update(gather_data["resolved"])

    for sls_ref, high_state in gather_data.get("state", {}).items():
        hub.idem.sls_source.init.decls(name, high_state, sls_ref)
        if hub.idem.RUNS[name]["errors"]:
            return

        # Check for duplicate state ids
        # This may occur when an SLS includes other SLSs -- the renderer can't catch these kinds of duplicates
        duplicates = set(high_state).intersection(hub.idem.RUNS[name]["high"])
        if duplicates:
            hub.idem.RUNS[name]["errors"].append(
                f"Duplicate state declarations found in SLS tree: {' '.join(duplicates)}"
            )
            return

        # Fire an event with the new high data
        await hub.idem.event.put(
            profile="idem-high",
            body=high_state,
            tags={"type": "state-high-data", "ref": "idem.sls_source.init.gather"},
        )
        hub.idem.RUNS[name]["high"].update(high_state)


def decls(
    hub,
    name: str,
    state: Dict[str, Any],
    sls_ref: str,
):
    """
    Resolve state formatting and data insertion

    :param hub:
    :param name: The state run name
    :param state: A rendered block from the sls
    :param sls_ref: A reference to another sls within the given sources
    """
    for id_ in hub.idem.resolve.init.iter(state):
        if not isinstance(state[id_], Dict):
            if isinstance(state[id_], str):
                # Is this is a short state, it needs to be padded
                if "." in state[id_]:
                    comps = state[id_].split(".")
                    state_ref = ".".join(comps[:-1])
                    state[id_] = {
                        "__sls__": sls_ref,
                        state_ref: [comps[-1]],
                    }
                    continue

            hub.idem.RUNS[name]["errors"].append(
                f"ID {id_} in SLS {sls_ref} is not a dictionary"
            )
            continue

        skeys = set()
        for key in list(state[id_]):
            if key.startswith("_"):
                continue
            if not isinstance(state[id_][key], List):
                continue
            if "." in key:
                comps = key.split(".")
                # Idem doesn't support state files such as:
                #
                #     /etc/redis/redis.conf:
                #       file.managed:
                #         - user: redis
                #         - group: redis
                #         - mode: 644
                #       file.comment:
                #           - regex: ^requirepass
                ref = ".".join(comps[:-1])
                if ref in skeys:
                    hub.idem.RUNS[name]["errors"].append(
                        f'ID "{id_}" in SLS "{sls_ref}" contains multiple state declarations from the same sub: {ref}'
                    )
                    continue
                state[id_][ref] = state[id_].pop(key)
                state[id_][ref].append(comps[-1])
                skeys.add(ref)
                continue
            skeys.add(key)

        if "__sls__" not in state[id_]:
            state[id_]["__sls__"] = sls_ref


def normalize(hub, path: str) -> str:
    """
    Transform a file path to a normalized SLS ref

    - Replace dots with \b and replace / with .
    - The existing dots may be part of a file name

    I.E
    "foo.bar/baz.bop.sls" will become "foo\bbar.baz\bbop"
    """
    p = pathlib.PurePosixPath(path)
    if "\b" in str(p) or "/" not in str(p):
        # We already have null characters or no "/"s in a pure posix path -- we can assume it is already normalized
        return path
    last_ref_name = p.stem if p.suffix == ".sls" else p.name
    normalized_ref = (
        str(pathlib.PurePosixPath(p.parent).joinpath(last_ref_name))
        # Double backs will temporarily be replaced with a backspace \0
        .replace("../", "\0")
        .replace(".", "\b")
        .replace("/", ".")
        # Swap the null character for a regular "."
        .replace("\0", ".")
    )
    return normalized_ref


def de_normalize(hub, path: str) -> str:
    """
    Transform an sls ref back into a file path for lookups

    I.E
    "foo\bbar.baz\bbop" will become "foo.bar/baz.bop"
    """
    if "/" in path:
        # Already denormalized
        return path

    # Replace leading dots with ../
    # I.E. ..ref.nest with ../../ref/nest
    match = re.match(r"^(\.*)(.*)$", path)
    levels, ref = match.groups()
    prepend = ""

    for _ in levels:
        # For each "." add a "../" to the de-normalized path
        prepend = f"../{prepend}"

    denormalized = ref.replace(".", "/").replace("\b", ".")

    return prepend + denormalized



from dataclasses import dataclass
from typing import ByteString
from typing import Tuple


async def sig_cache(
    hub, ctx, protocol: str, source: str, location: str
) -> Tuple[str, ByteString]:
    # Return a unique string as an sls file reference and sls content in ByteString
    ...


@dataclass
class CacheTarget:
    path: str
    value: bytes

    def __bool__(self):
        return bool(self.value)

    def __iter__(self):
        yield self.path
        yield self.value


async def post_cache(hub, ctx):
    # Validate the return
    if not ctx.ret:
        return CacheTarget("", b"")
    path, value = ctx.ret
    return CacheTarget(path, value)



from typing import Any
from typing import Dict


def sig_check(
    hub,
    name: str,
    ctx: Dict[str, Any],
    condition: Any,
    reqret: Dict[str, Any],
    chunk: Dict[str, Any],
    managed_state: Dict[str, Any],
    execution_seq: Dict,
) -> Dict[str, Any]:
    ...



def is_state_has_sls_run(hub, chunk):
    """
    Determine the state referred by chunk has flag __IS_SLS_RUN__ True
    """
    try:
        return getattr(hub.states[chunk["state"]], "__IS_SLS_RUN__")
    except Exception:
        return False



"""
Plugin for local state management
"""
import os
import pathlib
from typing import Any
from typing import Dict

import aiofiles


def __init__(hub):
    hub.esm.local.ACCT = ["esm.local"]


async def get_state(hub, ctx) -> Dict[str, Any]:
    """
    Read this context's state from it's cache_file

    :param hub:
    :param ctx:
    :return:
    """
    cache_file = _cache_file(ctx)
    async with aiofiles.open(cache_file, "rb+") as fh:
        data = await fh.read()
        if not data:
            return {}
        state = hub.serial[ctx.acct.serial_plugin].load(data)
        return state


async def set_state(hub, ctx, state: Dict[str, Any]):
    """
    Write the state to this context's cache_file

    :param hub:
    :param ctx:
    :param state:
    :return:
    """
    cache_file = _cache_file(ctx)
    async with aiofiles.open(cache_file, "wb+") as fh:
        data: bytes = hub.serial[ctx.acct.serial_plugin].dump(state)
        await fh.write(data)


async def enter(hub, ctx):
    """
    Verify that only one instance of run_name exists for the current user
    """
    pid_file = ctx.acct.cache_dir / "esm" / f"{ctx.acct.run_name}.pid"
    pid_file.parent.mkdir(exist_ok=True, parents=True)

    # Enter the context of the pid_file
    pid = await hub.esm.local.verify(pid_file)
    if pid:
        # A process with a different PID is already running for this run_name
        raise ChildProcessError(
            f"idem run '{ctx.acct.run_name}' is already active in process: {pid}"
        )

    # Write the current pid to the pid_file
    pid_file.parent.mkdir(exist_ok=True, parents=True)
    async with aiofiles.open(pid_file, "w+") as fh:
        pid = str(os.getpid())
        await fh.write(pid)

    return pid_file


async def exit_(hub, ctx, handle: pathlib.Path, exception: Exception):
    if exception:
        hub.log.error(f"{exception.__class__.__name__}: {exception}")
        return
    if handle.exists():
        handle.unlink()


async def verify(hub, pid_file: pathlib.Path) -> int or None:
    """
    If a pid is in the given file and is running, then return it.
    """
    if not pid_file.exists():
        # There is no existing pid_file, nothing to check
        return None

    async with aiofiles.open(pid_file, "r") as fh:
        contents = await fh.read()

    try:
        old_pid = int(contents)
    except (ValueError, TypeError):
        hub.log.error(f"Invalid pid_file '{fh.name}': {contents}")
        # There's no valid running pid, just return
        return None

    if hub.esm.local.exists(old_pid):
        # The pid_file has a pid in it that is still active, return it
        return old_pid


def exists(hub, pid: int) -> bool:
    """
    Check if a process exists with the given pid
    """
    try:
        # Send a no-op signal to the process
        os.kill(pid, 0)
        # If it was successful, the pid exists
        return True
    except OSError:
        # If there was an error, the pid does not exist
        return False


def _cache_file(ctx) -> pathlib.Path:
    """
    Return the Path to the cache file used for this ctx
    """
    cache_file: pathlib.Path = (
        ctx.acct.cache_dir
        / "esm"
        / "local"
        / f"{ctx.acct.run_name}.{ctx.acct.serial_plugin}"
    )
    # Ensure that the parent directories exist
    cache_file.parent.mkdir(parents=True, exist_ok=True)
    # Ensure that the file exists
    cache_file.touch()
    return cache_file



from typing import Any
from typing import Dict


async def sig_loop(
    hub,
    pending_plugin: str,
    name: str,
    apply_kwargs: Dict[str, Any],
):
    """
    Validate the signature of the reconciliation loop function
    """
    ...


def post_loop(hub, ctx):
    """
    Conform the return structure of every loop function to this format.
    """
    assert isinstance(ctx.ret["re_runs_count"], int)
    assert isinstance(ctx.ret["require_re_run"], bool)
    return ctx.ret



def stage(hub, name: str):
    """
    Take the highdata and convert argument references to arg_bind requisites
    """
    high, errors = hub.idem.ccomps.arg_bind.reconcile(hub.idem.RUNS[name]["high"])
    hub.idem.RUNS[name]["high"] = high
    hub.idem.RUNS[name]["errors"] = errors



import os
import tempfile

DEFAULT_ROOT_DIR = os.path.abspath(os.sep)
# Patch to change the default tempdir on Windows
if os.name == "nt":
    DEFAULT_CACHE_DIR = os.path.join(tempfile.gettempdir(), "cache", "idem")
else:
    DEFAULT_CACHE_DIR = "/var/cache/idem"


CLI_CONFIG = {
    "config": {"options": ["-c"], "subcommands": ["_global_"]},
    # Describe Options
    "desc_glob": {
        "display_priority": 0,
        "positional": True,
        "subcommands": ["describe", "refresh"],
        "help": "A glob to match the refs that should be ran",
    },
    "no_progress_bar": {
        "help": "Do not show a progress bar",
        "action": "store_false",
        "metavar": "progress",
        "dest": "progress",
        "subcommands": ["describe", "refresh", "state"],
        "ex_group": "progress",
    },
    "progress": {
        "action": "store_true",
        "subcommands": ["describe", "refresh", "state"],
        "ex_group": "progress",
    },
    "hard_fail": {
        "action": "store_true",
        "subcommands": ["describe", "refresh"],
        "help": "Hard stop the first time an error is raised",
    },
    "filter": {
        "type": str,
        "subcommands": ["describe"],
        "help": "A JMES search path",
    },
    # Exec Options
    "exec_func": {
        "display_priority": 0,
        "positional": True,
        "subcommands": ["exec"],
        "help": "The execution function to run by it's reference on the hub",
    },
    "exec_args": {
        "display_priority": 1,
        "positional": True,
        "nargs": "*",
        "render": "cli",
        "subcommands": ["exec"],
    },
    # State and validate options
    "run_name": {
        "subcommands": ["state", "exec", "describe", "validate", "refresh", "restore"],
    },
    "sls_sources": {"nargs": "*", "subcommands": ["state", "validate"]},
    "param_sources": {"nargs": "*", "subcommands": ["state", "describe"]},
    "test": {"options": ["-t"], "action": "store_true", "subcommands": ["state"]},
    "tree": {
        "options": ["-T"],
        "subcommands": ["state", "validate"],
    },
    "cache_dir": {"subcommands": ["state", "validate", "refresh", "restore"]},
    "root_dir": {"subcommands": ["state", "validate", "refresh", "restore"]},
    "params": {"nargs": "*", "subcommands": ["state", "describe"]},
    "group": {"subcommands": ["state"]},
    "invert": {
        "action": "store_true",
        "subcommands": ["state", "validate"],
        "help": "Inverts the action taken during state execution. For example, absent in case of present and present in case of absent.",
    },
    "render": {
        "subcommands": ["state", "validate"],
    },
    "runtime": {
        "subcommands": ["state"],
    },
    "reconciler": {
        "options": ["-r", "-R"],
        "subcommands": ["state"],
    },
    "pending": {
        "options": ["-p", "-P"],
        "subcommands": ["state"],
    },
    "output": {
        "source": "rend",
        "subcommands": [
            "exec",
            "state",
            "describe",
            "decrypt",
            "validate",
            "refresh",
            "doc",
        ],
    },
    "sls": {"positional": True, "nargs": "*", "subcommands": ["state", "validate"]},
    "target": {"subcommands": ["state"]},
    "esm_plugin": {"subcommands": ["state", "validate", "refresh", "restore"]},
    "esm_profile": {"subcommands": ["state", "validate", "refresh", "restore"]},
    # ACCT options
    "input_file": {
        "source": "acct",
        "positional": True,
        "subcommands": ["encrypt", "decrypt"],
    },
    "output_file": {
        "subcommands": ["encrypt"],
        "source": "acct",
    },
    "acct_file": {
        "source": "acct",
        "subcommands": ["state", "exec", "describe", "validate", "refresh", "restore"],
    },
    "acct_key": {
        "source": "acct",
        "subcommands": [
            "state",
            "exec",
            "describe",
            "encrypt",
            "decrypt",
            "validate",
            "refresh",
            "restore",
        ],
    },
    "acct_profile": {
        "subcommands": ["state", "exec", "describe", "validate", "refresh"],
    },
    "crypto_plugin": {
        "source": "acct",
        "subcommands": ["encrypt", "decrypt"],
    },
    "render_pipe": {
        "source": "acct",
        "subcommands": ["encrypt", "decrypt"],
    },
    "extras": {
        "source": "acct",
        "subcommands": ["state", "exec", "describe", "validate", "refresh"],
    },
    "overrides": {
        "source": "acct",
        "subcommands": ["state", "exec", "describe", "validate", "refresh"],
    },
    # EVBUS options
    "serialize_plugin": {
        "source": "evbus",
        "subcommands": ["exec", "describe", "state", "validate"],
    },
    # RESTORE options
    "esm_cache_file": {"positional": True, "subcommands": ["restore"]},
    # DOC options
    "ref": {
        "type": str,
        "help": "The reference on the hub to show",
        "source": "pop_tree",
        "positional": True,
        "subcommands": ["doc"],
        "nargs": "?",
    },
    "disable_jinja_sandbox": {
        "dest": "enable_jinja_sandbox",
        "subcommands": ["state", "validate"],
        "action": "store_false",
    },
    "jinja_sandbox_safe_hub_refs": {
        "source": "rend",
        "subcommands": ["state", "validate"],
        "nargs": "*",
    },
}

CONFIG = {
    "config": {
        "default": None,
        "help": "Load extra options from a .yaml configuration file onto hub.OPT",
    },
    "esm_plugin": {
        "default": "local",
        "help": "The esm (enforced state manager) plugin to use with a state run",
    },
    "esm_serial_plugin": {
        "default": "msgpack",
        "help": "The serial plugin used to store ESM data",
    },
    "esm_keep_cache": {
        "default": False,
        "help": "Do not delete the local ESM cache after it has been flushed to the ESM plugin",
    },
    "esm_profile": {
        "default": "default",
        "help": "The acct profile to use with the enforced state manager",
    },
    "run_name": {
        "default": "cli",
        "help": "A name for this run, this is used in internal tracking and to organize events",
    },
    "sls_sources": {
        "default": [],
        "help": "list off the sources that should be used for gathering sls files and data",
    },
    "param_sources": {
        "default": [],
        "help": "list off the sources that should be used for gathering parameter sls files",
    },
    "test": {
        "default": False,
        "help": "Set the idem run to execute in test mode. No changes will be made, idem will only detect if changes will be made in a real run.",
    },
    "tree": {
        "default": "",
        "help": "The directory containing sls files",
    },
    "params": {
        "default": [],
        "help": "The location of param.sls file containing parameter definitions",
    },
    "cache_dir": {
        "default": DEFAULT_CACHE_DIR,
        "help": "The location to use for the cache directory relative to the root_dir",
    },
    "root_dir": {
        "default": DEFAULT_ROOT_DIR,
        "help": f'The root directory to run idem from. By default it will be "{DEFAULT_ROOT_DIR}" or in the case of running as non-root it is set to <HOMEDIR>{os.sep}.idem',
    },
    "render": {
        "default": "jinja|yaml|replacements",
        "help": "The render pipe to use, this allows for the language to be specified",
    },
    "runtime": {
        "default": "parallel",
        "help": "Select which execution runtime to use. Options: 'parallel' (default), 'serial'.",
    },
    "sls": {
        "default": [],
        "help": "A space delimited list of sls refs to execute",
    },
    "exec": {
        "default": "",
        "help": "The name of an execution function to execute",
    },
    "exec_args": {
        "default": [],
        "help": "Arguments to pass to the named execution function",
    },
    "acct_profile": {
        "os": "ACCT_PROFILE",
        "help": "The profile to use when when calling exec modules and states",
        "default": "default",
    },
    "reconciler": {
        "default": "basic",
        "help": "The reconciler plugin to use, 'basic' by default",
    },
    "pending": {
        "default": "default",
        "help": "The pending plugin to use within the reconciler, 'default' by default",
    },
    "progress": {
        "default": True,
        "help": "Show a progress bar",
    },
    "progress_options": {
        # It is an overhead to refresh tqdm progress bar frequently. Default is 1 second with this configuration
        # This optimization reduces overhead on the system.
        "default": {"mininterval": 1},
        "help": "Keyword options to pass to the tqdm progress bar constructor",
        "type": dict,
    },
    "progress_plugin": {
        "default": "tqdm",
        "help": "The progress bar plugin to use",
        "type": dict,
    },
    "group": {
        "default": "number",
        "help": "The group plugins, separated by pipes, to pass the data through before it gets rendered by outputters",
    },
    "target": {
        "default": "",
        "help": "Specify a single resource declaration ID. The resource and all its dependencies will be executed.",
    },
    "enable_jinja_sandbox": {
        "source": "rend",
        "default": True,
        "help": "Enable sandboxed environment for Jinja rendering. Jinja sandboxing is enabled by default.",
    },
    "jinja_sandbox_safe_hub_refs": {
        "source": "rend",
        "default": ["exec.*", "idem.arg_bind.*"],
        "help": "Hub reference paths that should be allowed for Jinja rendering in the sandboxed environment. "
        "hub.exec.* amd hub.idem.arg_bind.* are enabled by default.",
    },
}

SUBCOMMANDS = {
    "encrypt": {
        "desc": "Use the acct subsystem to encrypt data",
        "help": "Use the acct subsystem to encrypt data",
        "source": "acct",
    },
    "decrypt": {
        "desc": "Use the acct subsystem to decrypt data",
        "help": "Use the acct subsystem to decrypt data",
        "source": "acct",
    },
    "state": {
        "desc": "Execute a specific state file or reference",
        "help": "Commands to run idempotent states",
    },
    "exec": {
        "desc": "Execute a specific execution routine",
        "help": "Commands to run execution routines",
    },
    "describe": {
        "desc": "Get an SLS representation of the account",
        "help": "Commands to run description routines",
    },
    "validate": {
        "desc": "Validate the given SLS tree and return the internal working data for the tree",
        "help": "Commands to validate SLS files",
    },
    "refresh": {
        "desc": "Update enforced state management with described resources",
        "help": "Update enforced state management with described resources",
    },
    "restore": {
        "desc": "Restore the enforced managed state from a cache file",
        "help": "Restore the enforced managed state from a cache file",
    },
    "doc": {
        "desc": "Print documentation from code",
        "help": "Print documentation from code",
    },
}
DYNE = {
    "esm": ["esm"],
    "source": ["source"],
    "idem": ["idem"],
    "exec": ["exec"],
    "log": ["log"],
    "states": ["states"],
    "tool": ["tool"],
    "output": ["output"],
    "reconcile": ["reconcile"],
    "acct": ["acct"],
    "group": ["group"],
}



from typing import Any
from typing import Dict


def sig_run(
    hub,
    seq: Dict[int, Dict[str, Any]],
    low: Dict[str, Any],
    running: Dict[str, Any],
    options: Dict[str, Any],
) -> Dict[int, Dict[str, Any]]:
    ...



from typing import Any
from typing import Dict


def apply(hub, data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Sort the output by the total seconds of the state's run time
    """
    ret = {}
    for tag in sorted(data, key=lambda k: data[k].get("total_seconds", 0)):
        ret[tag] = data[tag]
    return ret
